import asyncio
import datetime

import typer

from ffun.api.entities import GetLastEntriesResponse
from ffun.api.http_handlers import _external_entries
from ffun.application.application import with_app
from ffun.cli.commands import cleaner  # noqa: F401
from ffun.cli.commands import metrics  # noqa: F401
from ffun.cli.commands import processors_quality  # noqa: F401
from ffun.cli.commands import user_settings  # noqa: F401
from ffun.core import logging
from ffun.domain.domain import new_user_id
from ffun.feeds_links import domain as fl_domain
from ffun.library import domain as l_domain

logger = logging.get_module_logger()

cli_app = typer.Typer()


async def _profile_get_last_entries() -> GetLastEntriesResponse:
    backup_time_str = "2024-11-13 22:58:46.876683"
    backup_time = datetime.datetime.fromisoformat(backup_time_str)

    user_id = new_user_id()  # Replace this with the real user id

    period = datetime.datetime.now() - backup_time + datetime.timedelta(days=3)
    max_returned_entries = 10000

    linked_feeds = await fl_domain.get_linked_feeds(user_id)

    linked_feeds_ids = [link.feed_id for link in linked_feeds]

    entries = await l_domain.get_entries_by_filter(
        feeds_ids=linked_feeds_ids, period=period, limit=max_returned_entries
    )

    external_entries, tags_mapping = await _external_entries(
        entries, with_body=False, user_id=user_id, min_tag_count=0, include_tags=True
    )

    logger.info("almost_finishied", entries_number=len(external_entries), tags_number=len(tags_mapping))

    return GetLastEntriesResponse(entries=external_entries, tagsMapping=tags_mapping)


async def run_profile() -> None:
    async with with_app():
        logger.info("experiment_started")

        # insert profiled call here

        logger.info("experiment_finished")


@cli_app.command()
def profile() -> None:
    asyncio.run(run_profile())



import asyncio
import sys

from tabulate import tabulate

from ffun.application.application import with_app
from ffun.cli.application import app
from ffun.librarian.background_processors import processors
from ffun.librarian.domain import count_failed_entries, move_failed_entries_to_processor_queue


async def run_failed_entries_count() -> None:
    async with with_app():
        failed_entries_count = await count_failed_entries()

    table = []

    for processor_info in processors:
        table.append(
            [processor_info.id, processor_info.processor.name, failed_entries_count.get(processor_info.id, 0)]
        )

    sys.stdout.write(tabulate(table, headers=["processor id", "processor name", "failed entries"], tablefmt="grid"))


@app.command()
def failed_entries_count() -> None:
    asyncio.run(run_failed_entries_count())


async def run_failed_enties_move_to_queue(processor_id: int, limit: int) -> None:
    if not any(processor_info.id == processor_id for processor_info in processors):
        raise ValueError(f"Processor with id {processor_id} not found")

    async with with_app():
        await move_failed_entries_to_processor_queue(processor_id, limit=limit)
        await count_failed_entries()


@app.command()
def failed_entries_move_to_queue(processor_id: int, limit: int) -> None:
    asyncio.run(run_failed_enties_move_to_queue(processor_id, limit))



from typing import Any, Sequence

from ffun.core import logging
from ffun.librarian import errors
from ffun.librarian.entities import TagsExtractor, TextCleaner
from ffun.librarian.processors import base
from ffun.library.entities import Entry
from ffun.llms_framework import errors as llmsf_errors
from ffun.llms_framework.domain import call_llm, search_for_api_key
from ffun.llms_framework.entities import (
    ChatResponse,
    LLMCollectionApiKey,
    LLMConfiguration,
    LLMGeneralApiKey,
    LLMProvider,
)
from ffun.llms_framework.providers import llm_providers
from ffun.ontology.entities import ProcessorTag

logger = logging.get_module_logger()


class Processor(base.Processor):
    __slots__ = (
        "llm_config",
        "llm_provider",
        "entry_template",
        "text_cleaner",
        "tag_extractor",
        "collections_api_key",
        "general_api_key",
    )

    def __init__(  # noqa
        self,
        llm_provider: LLMProvider,
        llm_config: LLMConfiguration,
        entry_template: str,
        text_cleaner: TextCleaner,
        tag_extractor: TagsExtractor,
        collections_api_key: LLMCollectionApiKey | None,
        general_api_key: LLMGeneralApiKey | None,
        **kwargs: Any
    ):
        super().__init__(**kwargs)
        self.llm_config = llm_config
        self.llm_provider = llm_providers.get(llm_provider).provider
        self.entry_template = entry_template
        self.text_cleaner = text_cleaner
        self.tag_extractor = tag_extractor

        self.collections_api_key = collections_api_key
        self.general_api_key = general_api_key

    async def process(self, entry: Entry) -> list[ProcessorTag]:
        dirty_text = self.entry_template.format(entry=entry)

        cleaned_text = self.text_cleaner(dirty_text)

        requests = self.llm_provider.prepare_requests(self.llm_config, cleaned_text)

        api_key_usage = await search_for_api_key(
            llm=self.llm_provider,
            llm_config=self.llm_config,
            entry=entry,
            requests=requests,
            collections_api_key=self.collections_api_key,
            general_api_key=self.general_api_key,
        )

        if api_key_usage is None:
            raise errors.SkipEntryProcessing(message="no api key found")

        try:
            responses = await call_llm(
                llm=self.llm_provider, llm_config=self.llm_config, api_key_usage=api_key_usage, requests=requests
            )
        except llmsf_errors.TemporaryError as e:
            raise errors.TemporaryErrorInProcessor(message=str(e)) from e

        return self.extract_tags(responses)

    def extract_tags(self, responses: Sequence[ChatResponse]) -> list[ProcessorTag]:
        raw_tags = set()
        tags: list[ProcessorTag] = []

        for response in responses:
            raw_tags.update(self.tag_extractor(response.response_content()))

        for raw_tag in raw_tags:
            tags.append(ProcessorTag(raw_uid=raw_tag))

        return tags



from typing import Any, Iterable

from ffun.domain.entities import UserId
from ffun.user_settings import operations
from ffun.user_settings.entities import UserSettings
from ffun.user_settings.values import SettingsRegister, user_settings


async def save_setting(user_id: UserId, kind: int, value: Any, register: SettingsRegister = user_settings) -> None:
    value_to_save = register.get(kind).type.serialize(value)

    await operations.save_setting(user_id, kind, value_to_save)


def _full_settings(values: dict[int, Any], kinds: Iterable[int], register: SettingsRegister) -> UserSettings:
    result = {}

    for kind in kinds:
        if not register.has(kind):
            continue

        entity = register.get(kind)

        if kind in values:
            result[entity.key] = entity.type.deserialize(values[kind])
            continue

        result[entity.key] = entity.default

    return result  # type: ignore


async def load_settings(
    user_id: UserId, kinds: Iterable[int], register: SettingsRegister = user_settings
) -> UserSettings:
    values = await operations.load_settings_for_users([user_id], kinds)

    loaded_settings = values[user_id] if user_id in values else {}

    return _full_settings(loaded_settings, kinds, register=register)


async def load_settings_for_users(
    user_ids: Iterable[UserId], kinds: Iterable[int], register: SettingsRegister = user_settings
) -> dict[UserId, UserSettings]:
    values = await operations.load_settings_for_users(user_ids, kinds)

    return {user_id: _full_settings(user_values, kinds, register=register) for user_id, user_values in values.items()}


async def get_users_with_setting(kind: int, value: Any, register: SettingsRegister = user_settings) -> set[UserId]:
    value_to_find = register.get(kind).type.serialize(value)

    return await operations.get_users_with_setting(kind, value_to_find)


remove_setting_for_all_users = operations.remove_setting_for_all_users


async def remove_deprecated_settings(register: SettingsRegister = user_settings) -> None:
    all_kinds = await operations.find_all_kinds()

    for kind in all_kinds:
        if register.has(kind):
            continue

        await remove_setting_for_all_users(kind)



import asyncio
import contextlib
from typing import AsyncGenerator

import fastapi
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import ORJSONResponse

from ffun.api import http_handlers as api_http_handlers
from ffun.application import errors
from ffun.application import utils as app_utils
from ffun.application.settings import settings
from ffun.auth import supertokens as st
from ffun.auth.settings import AuthMode
from ffun.auth.settings import settings as auth_settings
from ffun.core import logging, middlewares, postgresql, sentry
from ffun.domain.urls import initialize_tld_cache
from ffun.feeds_collections.collections import collections
from ffun.loader import domain as l_domain

logger = logging.get_module_logger()


def initialize_user_settings() -> None:
    logger.info("initialize_user_settings")
    import ffun.application.user_settings  # noqa: F401


@contextlib.asynccontextmanager
async def use_postgresql() -> AsyncGenerator[None, None]:
    logger.info("initialize_postgresql")
    await postgresql.prepare_pool(
        name="ffun_pool",
        dsn=settings.postgresql.dsn,
        min_size=settings.postgresql.pool_min_size,
        max_size=settings.postgresql.pool_max_size,
        timeout=settings.postgresql.pool_timeout,
        num_workers=settings.postgresql.pool_num_workers,
        max_lifetime=settings.postgresql.pool_max_lifetime,
    )
    logger.info("postgresql_initialized")

    refresher = asyncio.create_task(
        postgresql.pool_refresher(settings.postgresql.pool_check_period), name="poll_refresher"
    )

    try:
        await postgresql.execute("""SELECT 1""")
    except Exception as e:
        raise errors.CanNotAccessPostgreSQL() from e

    try:
        yield
    finally:
        logger.info("deinitialize_postgresql")

        refresher.cancel()

        await postgresql.destroy_pool()

        logger.info("postgresql_deinitialized")


@contextlib.asynccontextmanager
async def use_api(app: fastapi.FastAPI) -> AsyncGenerator[None, None]:
    logger.info("api_enabled")
    app.include_router(api_http_handlers.router)

    logger.info("api_initialized")

    yield

    logger.info("api_deinitialized")


@contextlib.asynccontextmanager
async def use_sentry() -> AsyncGenerator[None, None]:
    logger.info("sentry_enabled")

    sentry.initialize(
        dsn=settings.sentry.dsn,
        sample_rate=settings.sentry.sample_rate,
        environment=settings.environment,
    )

    logger.info("sentry_initialized")

    yield

    logger.info("sentry_disabled")


def smart_url(domain: str, port: int) -> str:
    if port == 80:
        return f"http://{domain}"

    if port == 443:
        return f"https://{domain}"

    return f"http://{domain}:{port}"


def create_app() -> fastapi.FastAPI:  # noqa: CCR001
    logging.initialize(use_sentry=settings.enable_sentry)

    initialize_tld_cache()

    logger.info("create_app")

    @contextlib.asynccontextmanager
    async def lifespan(app: fastapi.FastAPI) -> AsyncGenerator[None, None]:
        async with contextlib.AsyncExitStack() as stack:
            await stack.enter_async_context(use_postgresql())

            if settings.enable_sentry:
                await stack.enter_async_context(use_sentry())

            if auth_settings.mode == AuthMode.supertokens:
                api_domain = smart_url(settings.app_domain, settings.api_port)
                website_domain = smart_url(settings.app_domain, settings.app_port)

                await stack.enter_async_context(
                    st.use_supertokens(
                        app_name=settings.app_name, api_domain=api_domain, website_domain=website_domain
                    )
                )

            if settings.enable_api:
                await stack.enter_async_context(use_api(app))

            await app.router.startup()

            await collections.initialize()

            initialize_user_settings()

            l_domain.initialize(user_agent=app_utils.user_agent())

            yield

            await app.router.shutdown()

    app = fastapi.FastAPI(
        lifespan=lifespan,
        docs_url=None,
        redoc_url=None,
        openapi_url=None,
        default_response_class=ORJSONResponse,
    )

    middlewares.initialize_error_processors(app)

    if auth_settings.mode == AuthMode.supertokens:
        st.add_middlewares(app)

    app.middleware("http")(middlewares.final_errors_middleware)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=settings.origins,
        allow_credentials=False,
        allow_methods=[],
        allow_headers=[],
    )

    app.middleware("http")(middlewares.request_measure_middleware)

    app.middleware("http")(middlewares.request_id_middleware)

    logger.info("app_created")

    return app


@contextlib.asynccontextmanager
async def with_app() -> AsyncGenerator[fastapi.FastAPI, None]:
    async with app.router.lifespan_context(app):
        yield app


app = create_app()



from ffun.core import errors


class Error(errors.Error):
    pass


class SkipEntryProcessing(Error):
    pass


class TemporaryErrorInProcessor(Error):
    pass


class CanNotSaveUnexistingPointer(Error):
    pass


class UnexpectedErrorInProcessor(Error):
    pass



from typing import Any

from ffun.librarian import errors
from ffun.library.entities import Entry
from ffun.ontology.entities import ProcessorTag


class Processor:
    __slots__ = ("_name",)

    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    async def process(self, entry: Entry) -> list[ProcessorTag]:
        raise NotImplementedError('You must implement "process" method in child class')


######################################
# Some processors for testing purposes
######################################


class AlwaysConstantProcessor(Processor):
    __slots__ = ("_tags",)

    def __init__(self, tags: list[str], **kwargs: Any):
        super().__init__(**kwargs)
        self._tags = tags

    async def process(self, entry: Entry) -> list[ProcessorTag]:
        return [ProcessorTag(raw_uid=tag) for tag in self._tags]


class AlwaysSkipEntryProcessor(Processor):
    async def process(self, entry: Entry) -> list[ProcessorTag]:
        raise errors.SkipEntryProcessing()


class AlwaysErrorProcessor(Processor):
    class CustomError(Exception):
        pass

    async def process(self, entry: Entry) -> list[ProcessorTag]:
        raise self.CustomError()


class AlwaysTemporaryErrorProcessor(Processor):
    async def process(self, entry: Entry) -> list[ProcessorTag]:
        raise errors.TemporaryErrorInProcessor()



import datetime
import functools
import pathlib

import pydantic
import pydantic_settings
import toml

from ffun.core import logging
from ffun.core.settings import BaseSettings
from ffun.llms_framework.entities import ModelInfo

logger = logging.get_module_logger()


_root = pathlib.Path(__file__).parent


class ModelInfos(pydantic.BaseModel):
    models: list[ModelInfo]


class Settings(BaseSettings):
    # TODO: add to documentation/README
    models_descriptions: pathlib.Path = _root / "fixtures" / "models.toml"

    key_quota_timeout: datetime.timedelta = datetime.timedelta(hours=1)
    key_broken_timeout: datetime.timedelta = datetime.timedelta(hours=1)

    @pydantic.computed_field  # type: ignore
    @functools.cached_property
    def models(self) -> list[ModelInfo]:
        data = toml.loads(self.models_descriptions.read_text())

        return ModelInfos(**data).models

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_LLMS_FRAMEWORK_")


settings = Settings()



import pydantic_settings

from ffun.core.settings import BaseSettings


class Settings(BaseSettings):
    max_entries_per_feed: int = 10000

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_LIBRARY_")


settings = Settings()



import os
from pathlib import Path

_excluded_suffixes = [".expected.json", "~", "#"]


feeds_fixtures_directory = Path(__file__).parent / "feeds_fixtures"


def feeds_fixtures_names() -> list[str]:
    files = os.listdir(feeds_fixtures_directory)

    return [filename for filename in files if not any(filename.endswith(suffix) for suffix in _excluded_suffixes)]



import asyncio

from ffun.cli.application import app
from ffun.core import migrations


async def run() -> None:
    await migrations.apply_all()


@app.command()
def migrate() -> None:
    asyncio.run(run())



from ffun.core import errors
from ffun.feeds import entities as f_entities


class Error(errors.Error):
    pass


class LoadError(Error):
    feed_error_code: f_entities.FeedError


class AllProxiesSuspended(Error):
    pass



from typing import Annotated, Literal

import pydantic

from ffun.core.entities import BaseEntity
from ffun.domain.entities import EntryId

EventsViewName = Literal["news", "rules", "public_collections", "settings", "collections", "discovery", "feeds"]


class NewsLinkOpened(BaseEntity):
    name: Literal["news_link_opened"]
    view: EventsViewName
    entry_id: EntryId


class NewsBodyOpened(BaseEntity):
    name: Literal["news_body_opened"]
    view: EventsViewName
    entry_id: EntryId


class SocialLinkClicked(BaseEntity):
    name: Literal["social_link_clicked"]
    view: EventsViewName
    link_type: Literal["api", "blog", "reddit", "discord", "github"]


class TagFilterStateChanged(BaseEntity):
    name: Literal["tag_filter_state_changed"]
    view: EventsViewName
    tag: str
    from_state: Literal["required", "excluded", "none"]
    to_state: Literal["required", "excluded", "none"]
    source: Literal["tags_filter", "entry_record", "rule_record"]


class SidebarStateChanged(BaseEntity):
    name: Literal["sidebar_state_changed"]
    view: EventsViewName
    sub_event: Literal["show", "hide"]
    source: Literal["top_sidebar_button"]


class UserUtm(BaseEntity):
    name: Literal["user_utm"]
    utm_source: str | None = None
    utm_medium: str | None = None
    utm_campaign: str | None = None


Event = Annotated[
    NewsLinkOpened | NewsBodyOpened | SocialLinkClicked | TagFilterStateChanged | SidebarStateChanged | UserUtm,
    pydantic.Field(discriminator="name"),
]



from ffun.core.tests.helpers import assert_compare_xml
from ffun.domain.urls import str_to_feed_url, url_to_uid
from ffun.feeds.entities import Feed
from ffun.parsers.entities import FeedInfo
from ffun.parsers.opml import create_opml, extract_feeds


class TestCreateOpml:

    def test(self, saved_feed: Feed, another_saved_feed: Feed) -> None:

        feeds = [saved_feed, another_saved_feed]
        feeds.sort(key=lambda feed: feed.title if feed.title is not None else "")

        content = create_opml(feeds)

        expected_content = f'<opml version="2.0"><head><title>Your subscriptions in feeds.fun</title></head><body><outline title="uncategorized" text="uncategorized"><outline title="{feeds[0].title}" text="{feeds[0].title}" type="rss" xmlUrl="{feeds[0].url}" /><outline title="{feeds[1].title}" text="{feeds[1].title}" type="rss" xmlUrl="{feeds[1].url}" /></outline></body></opml>'  # noqa: E501

        assert_compare_xml(content, expected_content.strip())


class TestExtractFeeds:

    def test(self, saved_feed: Feed, another_saved_feed: Feed) -> None:
        feeds = [saved_feed, another_saved_feed]
        feeds.sort(key=lambda feed: feed.title if feed.title is not None else "")

        content = create_opml(feeds)

        infos = extract_feeds(content)

        infos.sort(key=lambda info: info.title)

        assert infos[0] == FeedInfo(
            url=feeds[0].url,
            title=feeds[0].title if feeds[0].title is not None else "",
            description="",
            entries=[],
            uid=url_to_uid(feeds[0].url),
        )

        assert infos[1] == FeedInfo(
            url=feeds[1].url,
            title=feeds[1].title if feeds[1].title is not None else "",
            description="",
            entries=[],
            uid=url_to_uid(feeds[1].url),
        )

    def test_opml_with_no_head(self) -> None:
        content = """
<opml version="2.0">
<body>
  <outline text="Subscriptions" title="Subscriptions">
    <outline type="rss" xmlUrl='https://example.com/feed' />
  </outline>
</body>
</opml>
        """

        infos = extract_feeds(content)

        expected_url = str_to_feed_url("https://example.com/feed")

        assert len(infos) == 1
        assert infos[0] == FeedInfo(
            url=expected_url,
            title="",
            description="",
            entries=[],
            uid=url_to_uid(expected_url),
        )

    def test_opml_with_no_type_attribute(self) -> None:
        content = """
<opml version="2.0">
<head>
  <title>My Subscriptions</title>
</head>
<body>
  <outline text="Subscriptions" title="Subscriptions">
    <outline xmlUrl='https://example.com/feed' />
  </outline>
</body>
</opml>
        """

        infos = extract_feeds(content)

        expected_url = str_to_feed_url("https://example.com/feed")

        assert len(infos) == 1
        assert infos[0] == FeedInfo(
            url=expected_url,
            title="",
            description="",
            entries=[],
            uid=url_to_uid(expected_url),
        )

    def test_opml_with_wrong_type_case(self) -> None:
        content = """
<opml version="2.0">
<head>
  <title>My Subscriptions</title>
</head>
<body>
  <outline text="Subscriptions" title="Subscriptions">
    <outline type="RsS" xmlUrl='https://example.com/feed' />
  </outline>
</body>
</opml>
        """

        infos = extract_feeds(content)

        expected_url = str_to_feed_url("https://example.com/feed")

        assert len(infos) == 1
        assert infos[0] == FeedInfo(
            url=expected_url,
            title="",
            description="",
            entries=[],
            uid=url_to_uid(expected_url),
        )



import re

from slugify import slugify

from ffun.domain.entities import TagUid

DISALLOWED_CHARS_PATTERN = re.compile(r"[^-a-zA-Z0-9]+")


_encode_replacements = {
    "#": "-sharp-",  # c# -> c-sharp
    "+": "-plus-",  # c++ -> c-plus-plus
    ".": "-dot-",  # .net -> dot-net, example.com -> example-dot-com
}


_decode_replacements = {v.strip("-"): k for k, v in _encode_replacements.items()}


def _encode_special_characters(tag: str) -> str:
    for substring, replacement in _encode_replacements.items():
        tag = tag.replace(substring, replacement)

    return tag


def _decode_special_characters(tag: str) -> str:
    parts = tag.split("-")
    parts = [part for part in parts if part]

    result = []
    can_add_dash = False

    for part in parts:
        if part not in _decode_replacements:
            if can_add_dash:
                result.append("-")

            can_add_dash = True
            result.append(part)
            continue

        can_add_dash = False
        result.append(_decode_replacements[part])

    return "".join(result)


def normalize(tag: str) -> TagUid:
    tag = tag.lower()

    tag = _encode_special_characters(tag)

    return slugify(
        tag,
        entities=True,
        decimal=True,
        hexadecimal=True,
        max_length=0,
        word_boundary=False,
        save_order=True,
        separator="-",
        stopwords=(),
        regex_pattern=DISALLOWED_CHARS_PATTERN,  # type: ignore
        lowercase=True,
        replacements=(),
        allow_unicode=False,
    )


def verbose(tag: str) -> str:
    return _decode_special_characters(tag)



"""
add pure users table
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230527_01_soIr3-users-mapping"}


sql_create_users_table = """
CREATE TABLE u_users (
    id UUID PRIMARY KEY,
    created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP
);
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_users_table)

    sql = """
        INSERT INTO u_users (id, created_at)
        SELECT internal_id, MIN(created_at)
        FROM u_mapping
        GROUP BY internal_id
    """

    cursor.execute(sql)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE u_users")


steps = [step(apply_step, rollback_step)]



from ffun.domain.entities import UserId
from ffun.users import errors, operations
from ffun.users.entities import Service, User

add_mapping = operations.add_mapping
get_mapping = operations.get_mapping
count_total_users = operations.count_total_users
tech_move_user = operations.tech_move_user
get_user_external_ids = operations.get_user_external_ids
unlink_user = operations.unlink_user


async def get_or_create_user_id(service: Service, external_id: str) -> UserId:
    try:
        return await get_mapping(service, external_id)
    except errors.NoUserMappingFound:
        return await add_mapping(service, external_id)


async def get_or_create_user(service: Service, external_id: str) -> User:
    user_id = await get_or_create_user_id(service, external_id)

    return User(id=user_id)



"""
users-mapping
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__: set[str] = set()


sql_create_user_mapping_table = """
CREATE TABLE u_mapping (
    external_id TEXT NOT NULL,
    internal_id UUID NOT NULL,
    service_id INTEGER NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
    PRIMARY KEY (service_id, external_id)
);
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_user_mapping_table)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE u_mapping")


steps = [step(apply_step, rollback_step)]



import contextlib
import re
import unicodedata
from typing import Iterable, Iterator
from urllib.parse import quote_plus, unquote

import tldextract
from furl import furl
from orderedmultidict import omdict

from ffun.core import logging
from ffun.domain import errors
from ffun.domain.entities import AbsoluteUrl, FeedUrl, SourceUid, UnknownUrl, UrlUid

logger = logging.get_module_logger()


# ATTENTION: in case of modification, you MUST check:
#            - that the logic of dependent functions is not broken
#            - that the UIDs generation is not changed (check on backup)
#            - in case UIDs generation is changed, you MUST update all affected entities


RE_SCHEMA = re.compile(r"^(\w+):")


@contextlib.contextmanager
def check_furl_error() -> Iterator[None]:
    try:
        yield
    except ValueError as e:
        message = str(e)

        if "Invalid port" in message:
            raise errors.ExpectedFUrlError(message=message) from e

        if "Invalid host" in message:
            raise errors.ExpectedFUrlError(message=message) from e

        raise


def construct_f_url(url: UnknownUrl | AbsoluteUrl | str) -> furl | None:
    try:
        with check_furl_error():
            return furl(url)
    except errors.ExpectedFUrlError:
        return None


def initialize_tld_cache() -> None:
    logger.info("initializing_tld_cache")
    tldextract.extract("example.com")
    logger.info("tld_cache_initialized")


# ATTENTION: see note at the top of the file
def schema_supported(scheme: str | None) -> bool:
    if scheme in (None, "http", "https"):
        return True

    # TODO: When we import OPML files, we may encounter non-standard schemas
    #       such schemas are used by other services to mark special cases, like newsletter-> RSS
    #       we may want to support them in the future
    #
    # Most popular schema is "newsletter" (Inoreader, Feedbin, NewsBlur, Feedly, Readwise Reader, Omnivore, Matter)
    # but there may be others

    return False


# ATTENTION: see note at the top of the file
def _fix_classic_url_to_absolute(url: str) -> AbsoluteUrl | None:
    domain_part = url.split("/")[0]

    # simple protection from processing special domains
    if "." not in domain_part:
        return None

    # check if url has a proper domain
    if tldextract.extract(domain_part).suffix == "":
        return None

    f_url = construct_f_url(f"//{url}")

    if f_url is None:
        return None

    return normalize_classic_unknown_url(UnknownUrl(str(f_url)))


# ATTENTION: see note at the top of the file
def normalize_classic_unknown_url(url: UnknownUrl) -> AbsoluteUrl | None:  # noqa: CCR001
    url = UnknownUrl(url.strip())

    # check if url is parsable
    f_url = construct_f_url(url)

    if f_url is None:
        return None

    if f_url.path == "/":
        f_url.path = None

    if url.startswith("//"):
        return AbsoluteUrl(str(f_url))

    if url.startswith("./") or url.startswith("../"):
        return None

    if match := RE_SCHEMA.match(url):

        if not schema_supported(match.group(1)):
            return None

        return AbsoluteUrl(str(f_url))

    return _fix_classic_url_to_absolute(url)


# ATTENTION: see note at the top of the file
def is_full_url(url: UnknownUrl) -> bool:
    return normalize_classic_unknown_url(url) is not None


# it is a shortcut method for tests
def str_to_absolute_url(url: str) -> AbsoluteUrl:
    """Convert or raise Exception

    It is shortcut method mostly for tests.
    Use `normalize_classic_unknown_url` in the production code.
    """
    absolute_url = normalize_classic_unknown_url(UnknownUrl(url))

    if absolute_url is None:
        raise errors.UrlIsNotAbsolute(url=url)

    return absolute_url


# it is a shortcut method for tests
def str_to_feed_url(url: str) -> FeedUrl:
    return to_feed_url(str_to_absolute_url(url))


# ATTENTION: see note at the top of the file
def is_absolute_url(url: str) -> bool:
    """Check if the URL is absolute and is normalized"""
    return normalize_classic_unknown_url(UnknownUrl(url)) == url


# ATTENTION: see note at the top of the file
def adjust_classic_full_url(url: UnknownUrl, original_url: AbsoluteUrl | FeedUrl) -> AbsoluteUrl | None:
    fixed_url = normalize_classic_unknown_url(url)
    assert fixed_url is not None

    f_original_url = furl(original_url)
    f_url = furl(fixed_url)

    # own schema has priority over origin schema
    # we expect that owner of site (who specify urls) know what they are doing
    if not f_url.scheme:
        f_url.scheme = f_original_url.scheme

    return AbsoluteUrl(str(f_url))


# ATTENTION: see note at the top of the file
def adjust_classic_relative_url(url: UnknownUrl, original_url: AbsoluteUrl | FeedUrl) -> AbsoluteUrl | None:
    f_url = construct_f_url(original_url)

    if f_url is None:
        return None

    f_url.remove(query_params=True, fragment=True)

    try:
        with check_furl_error():
            f_url.join(url)
    except errors.ExpectedFUrlError:
        return None

    return AbsoluteUrl(str(f_url))


# ATTENTION: see note at the top of the file
def adjust_classic_url(url: UnknownUrl, original_url: AbsoluteUrl | FeedUrl) -> AbsoluteUrl | None:
    if is_full_url(url):
        return adjust_classic_full_url(url, original_url)

    return adjust_classic_relative_url(url, original_url)


def is_magnetic_url(url: UnknownUrl) -> bool:
    return url.startswith("magnet:")


def adjust_magnetic_url(url: UnknownUrl) -> AbsoluteUrl:
    return AbsoluteUrl(url)


# ATTENTION: see note at the top of the file
def adjust_external_url(url: UnknownUrl, original_url: AbsoluteUrl | FeedUrl) -> AbsoluteUrl | None:
    if is_magnetic_url(url):
        return adjust_magnetic_url(url)

    return adjust_classic_url(url, original_url)


# ATTENTION: see note at the top of the file
def url_to_uid(url: AbsoluteUrl | FeedUrl) -> UrlUid:
    # The goal of this function is to detect URLs that most likely (99.(9)%) point to the same resource
    # It normalizes and simplifies a URL according to heuristics
    # I.e. there is a small possibility that two different URLs will be normalized to the same uid
    #
    # For example, http://example.com/ and http://example.com can return different data
    # In reality, it is mostly impossible and is a sign of bug or hacking on the side of third-party service
    # => We could remove a schema from the resulting uid
    #
    # Some normalization rules are based on personal taste,
    # for example, there are multiple ways to encode a URL or to normalize Unicode
    #
    # The rules are based on the next heuristics:
    #
    # - readability is better than technical representation
    # - it is ok to loss some corner URL forms, unless there is an explicit request to support them

    normalized_url = url.lower().strip()

    normalized_url = unquote(normalized_url)  # unquote all
    normalized_url = quote_plus(normalized_url)  # quoute with replacing spaces with pluses
    normalized_url = unquote(normalized_url)  # unquote all again, but keep pluses

    normalized_url = unicodedata.normalize("NFC", normalized_url)

    url_object = furl(normalized_url)

    url_object.scheme = None
    url_object.port = None
    url_object.fragment = None

    url_object.query.params = omdict(sorted(url_object.query.params.allitems()))

    # Attention: we must not remove username:password from the url
    #            because it will create a vector for hacking by accessing private news of other users

    path = str(url_object.path)

    while "//" in path:
        path = path.replace("//", "/")

    if path and path[-1] == "/":
        path = path[:-1]

    url_object.path = path

    # unquote again because furl will quote all parts of the url
    resulted_url = unquote(str(url_object))

    # furl adds // at the beginning of the url in some cases
    if resulted_url.startswith("//"):
        resulted_url = resulted_url[2:]

    return UrlUid(resulted_url)


# ATTENTION: see note at the top of the file
def url_to_source_uid(url: AbsoluteUrl | FeedUrl) -> SourceUid:
    # Because some portals (Reddit, ArXiv) provide customizable feed URLs,
    # we could see the same news entry in different feeds
    # => we should track the entry's source not by feed but by the portal
    # that will help us to ensure the entry's uniqueness.

    normalized_url = unicodedata.normalize("NFC", url).lower().strip()

    url_object = furl(normalized_url)

    domain = url_object.host

    # TODO: move rules to settings

    if domain.startswith("www."):
        domain = domain[4:]

    if domain.endswith(".reddit.com"):
        # xxx.reddit.com domains are just the old GUI version of reddit.com, or API, or old RSS urls
        domain = "reddit.com"

    assert isinstance(domain, str)

    return SourceUid(domain)


def url_has_extension(url: AbsoluteUrl, expected_extensions: list[str]) -> bool:
    f_url = furl(url)

    if not f_url.path.segments:
        return "" in expected_extensions

    last_segment = f_url.path.segments[-1]

    if "." not in last_segment:
        return "" in expected_extensions

    extension = last_segment.rsplit(".")[-1].strip()

    return f".{extension}" in expected_extensions


def filter_out_duplicated_urls(urls: Iterable[AbsoluteUrl]) -> list[AbsoluteUrl]:
    seen = set()

    result = []

    for url in urls:
        uid = url_to_uid(url)

        if uid in seen:
            continue

        seen.add(uid)
        result.append(url)

    return result


def get_parent_url(url: AbsoluteUrl | FeedUrl) -> AbsoluteUrl | None:
    f_url = furl(url)

    if not f_url.path.segments or f_url.path == "/":
        return None

    f_url.remove(query_params=True, fragment=True)

    if f_url.path.segments[-1] == "":
        f_url.path.segments = f_url.path.segments[:-1]
    else:
        f_url.path.segments[-1] = ""

    return normalize_classic_unknown_url(UnknownUrl(str(f_url)))


def to_feed_url(url: AbsoluteUrl) -> FeedUrl:
    f_url = furl(url)

    f_url.fragment = None

    return FeedUrl(str(f_url))



"""
tags-properties
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230617_02_L0MmA-tags-relationship-processor-tracking"}

sql_create_tags_properties_table = """
CREATE TABLE o_tags_properties (
    id BIGSERIAL PRIMARY KEY,
    tag_id BIGINT REFERENCES o_tags(id),
    type SMALLINT NOT NULL,
    value TEXT NOT NULL,
    processor_id BIGINT NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW())
"""

sql_create_tags_properties_unique_index = """
CREATE UNIQUE INDEX idx_o_tags_properties_unique ON o_tags_properties(tag_id, type, processor_id)
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_tags_properties_table)
    cursor.execute(sql_create_tags_properties_unique_index)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE o_tags_properties")


steps = [step(apply_step, rollback_step)]



import uuid
from itertools import chain

import pytest
import pytest_asyncio
from pytest_mock import MockerFixture

from ffun.core import utils
from ffun.core.postgresql import execute
from ffun.domain.entities import UserId
from ffun.domain.urls import str_to_feed_url, url_to_source_uid, url_to_uid
from ffun.feeds import domain as f_domain
from ffun.feeds.entities import Feed
from ffun.feeds.tests import make as f_make
from ffun.feeds_links import domain as fl_domain
from ffun.library import domain as l_domain
from ffun.library.tests import make as l_make
from ffun.meta.domain import add_feeds, clean_orphaned_entries, clean_orphaned_feeds, remove_entries
from ffun.ontology import domain as o_domain
from ffun.ontology.entities import ProcessorTag
from ffun.parsers import entities as p_entities


class TestRemoveEntries:
    @pytest.mark.asyncio
    async def test_no_entries(self) -> None:
        await remove_entries([])

    @pytest.mark.asyncio
    async def test_success(
        self,
        loaded_feed: Feed,
        another_loaded_feed: Feed,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag],
    ) -> None:
        entries = await l_make.n_entries_list(loaded_feed, 3)
        another_entries = await l_make.n_entries_list(another_loaded_feed, 3)

        tag_a, tag_b, tag_c = three_processor_tags

        # fill feed 1
        await o_domain.apply_tags_to_entry(entry_id=entries[0].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(entry_id=entries[1].id, processor_id=fake_processor_id, tags=[tag_a])

        await o_domain.apply_tags_to_entry(
            entry_id=entries[1].id, processor_id=another_fake_processor_id, tags=[tag_b]
        )

        # fill feed 2
        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[1].id, processor_id=fake_processor_id, tags=[tag_a]
        )

        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[1].id, processor_id=another_fake_processor_id, tags=[tag_b]
        )

        await o_domain.apply_tags_to_entry(
            entry_id=another_entries[2].id, processor_id=another_fake_processor_id, tags=[tag_c]
        )

        await remove_entries([entries[0].id, another_entries[1].id, entries[2].id])

        loaded_entries = await l_domain.get_entries_by_ids(
            [entry.id for entry in entries] + [entry.id for entry in another_entries]
        )

        assert loaded_entries == {
            entries[0].id: None,
            entries[1].id: entries[1],
            entries[2].id: None,
            another_entries[0].id: another_entries[0],
            another_entries[1].id: None,
            another_entries[2].id: another_entries[2],
        }


class TestAddFeeds:

    @pytest.mark.asyncio
    async def test_no_feeds_to_add(self, internal_user_id: UserId) -> None:
        await add_feeds([], internal_user_id)

    @pytest.mark.asyncio
    async def test_add(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:

        urls = [str_to_feed_url(f"{uuid.uuid4().hex}.com") for _ in range(3)]

        feeds = [
            p_entities.FeedInfo(
                url=urls[0],
                title=uuid.uuid4().hex,
                description=uuid.uuid4().hex,
                entries=[],
                uid=url_to_uid(urls[0]),
            ),
            p_entities.FeedInfo(
                url=urls[1],
                title=uuid.uuid4().hex,
                description=uuid.uuid4().hex,
                entries=[],
                uid=url_to_uid(urls[1]),
            ),
            p_entities.FeedInfo(
                url=urls[2], title=uuid.uuid4().hex, description=uuid.uuid4().hex, entries=[], uid=url_to_uid(urls[2])
            ),
        ]

        await add_feeds(feeds[:2], internal_user_id)
        await add_feeds(feeds[1:], another_internal_user_id)

        links_1 = await fl_domain.get_linked_feeds(internal_user_id)
        links_2 = await fl_domain.get_linked_feeds(another_internal_user_id)

        feeds_1 = await f_domain.get_feeds([link.feed_id for link in links_1])
        feeds_2 = await f_domain.get_feeds([link.feed_id for link in links_2])

        assert len({feed.id for feed in feeds_1} & {feed.id for feed in feeds_2}) == 1

        assert {feed.url for feed in feeds_1} == {feed.url for feed in feeds[:2]}
        assert {feed.url for feed in feeds_2} == {feed.url for feed in feeds[1:]}

        source_uids = {url: url_to_source_uid(url) for url in urls}
        source_ids = await f_domain.get_source_ids(source_uids.values())

        for feed in chain(feeds_1, feeds_2):
            assert feed.source_id == source_ids[source_uids[feed.url]]


# test that everything is connected correctly
# detailed cases are covered in tests of functions called in clean_orphaned_entries
class TestCleanOrphanedEntries:

    @pytest_asyncio.fixture(autouse=True)
    async def cleanup_orphaned_entries(self) -> None:
        await execute("DELETE FROM l_orphaned_entries")

    @pytest.mark.asyncio
    async def test(self, loaded_feed: Feed) -> None:
        entries = await l_make.n_entries_list(loaded_feed, n=10)

        await l_domain.unlink_feed_tail(loaded_feed.id, offset=3)

        removed_1 = await clean_orphaned_entries(chunk=5)

        assert removed_1 == 5

        removed_2 = await clean_orphaned_entries(chunk=5)

        assert removed_2 == 2

        loaded_entries = await l_domain.get_entries_by_ids([entry.id for entry in entries])

        assert loaded_entries == {entry.id: entry for entry in entries[:3]} | {entry.id: None for entry in entries[3:]}


# test that everything is connected correctly
# detailed cases are covered in tests of functions called in clean_orphaned_entries
class TestCleanOrphanedFeeds:

    @pytest_asyncio.fixture(autouse=True)
    async def cleanup_orphaned_feeds(self) -> None:
        orphanes = await f_domain.get_orphaned_feeds(limit=10000, loaded_before=utils.now())

        for orphane_id in orphanes:
            await f_domain.tech_remove_feed(orphane_id)

    @pytest.mark.asyncio
    async def test_chunks(self) -> None:
        feeds = await f_make.n_feeds(10)

        for feed in feeds:
            await f_domain.mark_feed_as_orphaned(feed.id)

        assert await clean_orphaned_feeds(chunk=3) == 3
        assert await clean_orphaned_feeds(chunk=4) == 4
        assert await clean_orphaned_feeds(chunk=5) == 3

    @pytest.mark.asyncio
    async def test_all_logic_called(self, mocker: MockerFixture) -> None:
        feeds = await f_make.n_feeds(3)

        feeds.sort(key=lambda feed: feed.id)

        for feed in feeds:
            await f_domain.mark_feed_as_orphaned(feed.id)

        unlink_feed_tail_mock = mocker.patch("ffun.library.domain.unlink_feed_tail")
        tech_remove_feed_mock = mocker.patch("ffun.feeds.domain.tech_remove_feed")
        tech_remove_all_links = mocker.patch("ffun.feeds_links.domain.tech_remove_all_links")

        assert await clean_orphaned_feeds(chunk=100) == 3

        assert unlink_feed_tail_mock.call_args_list == [mocker.call(feed.id, offset=0) for feed in feeds]

        assert tech_remove_feed_mock.call_args_list == [mocker.call(feed.id) for feed in feeds]

        assert tech_remove_all_links.call_args_list == [mocker.call([feed.id for feed in feeds])]



import datetime
import uuid
from decimal import Decimal
from typing import Any

import pytest
from pytest_mock import MockerFixture

from ffun.application.resources import Resource as AppResource
from ffun.core.tests.helpers import assert_logs_has_business_event, capture_logs
from ffun.domain.datetime_intervals import month_interval_start
from ffun.domain.domain import new_user_id
from ffun.domain.entities import UserId
from ffun.feeds.entities import FeedId
from ffun.feeds_collections.collections import collections
from ffun.feeds_collections.entities import CollectionId
from ffun.feeds_links import domain as fl_domain
from ffun.llms_framework import errors
from ffun.llms_framework.entities import (
    APIKeyUsage,
    KeyStatus,
    LLMApiKey,
    LLMCollectionApiKey,
    LLMConfiguration,
    LLMGeneralApiKey,
    LLMProvider,
    LLMTokens,
    SelectKeyContext,
    USDCost,
    UserKeyInfo,
)
from ffun.llms_framework.keys_rotator import (
    _api_key_is_working,
    _choose_collections_key,
    _choose_general_key,
    _choose_user,
    _choose_user_key,
    _cost_points,
    _filter_out_users_for_whome_entry_is_too_old,
    _filter_out_users_with_overused_keys,
    _filter_out_users_with_wrong_keys,
    _filter_out_users_without_keys,
    _filters,
    _find_best_user_with_key,
    _get_candidates,
    _get_user_key_infos,
    _key_selectors,
    choose_api_key,
    use_api_key,
)
from ffun.llms_framework.provider_interface import ProviderInterface, ProviderTest
from ffun.llms_framework.providers import llm_providers
from ffun.resources import domain as r_domain
from ffun.resources import entities as r_entities
from ffun.user_settings import domain as us_domain

_llm_config = LLMConfiguration(
    model="test-model",
    system="some system prompt",
    max_return_tokens=LLMTokens(1017),
    text_parts_intersection=113,
    temperature=0.3,
    top_p=0.9,
    presence_penalty=0.5,
    frequency_penalty=0.75,
)


class TestAPIKeyIsWorking:
    @pytest.mark.asyncio
    async def test_is_working(self, fake_llm_provider: ProviderTest, fake_llm_api_key: LLMApiKey) -> None:
        fake_llm_provider.api_keys_statuses.set(fake_llm_api_key, KeyStatus.works)
        assert await _api_key_is_working(fake_llm_provider, _llm_config, fake_llm_api_key)

    @pytest.mark.parametrize(
        "status", [status for status in KeyStatus if status not in [KeyStatus.works, KeyStatus.unknown]]
    )
    @pytest.mark.asyncio
    async def test_guaranted_broken(
        self, fake_llm_provider: ProviderTest, status: KeyStatus, fake_llm_api_key: LLMApiKey
    ) -> None:
        fake_llm_provider.api_keys_statuses.set(fake_llm_api_key, status)
        assert not await _api_key_is_working(fake_llm_provider, _llm_config, fake_llm_api_key)

    @pytest.mark.parametrize("status, expected_result", [(KeyStatus.works, True), (KeyStatus.broken, False)])
    @pytest.mark.asyncio
    async def test_unknown(
        self,
        fake_llm_provider: ProviderTest,
        status: KeyStatus,
        expected_result: bool,
        mocker: MockerFixture,
        fake_llm_api_key: LLMApiKey,
    ) -> None:
        assert fake_llm_provider.api_keys_statuses.get(fake_llm_api_key) == KeyStatus.unknown

        mocker.patch("ffun.llms_framework.provider_interface.ProviderTest.check_api_key", return_value=status)

        assert await _api_key_is_working(fake_llm_provider, _llm_config, fake_llm_api_key) == expected_result


class TestFilterOutUsersWithWrongKeys:
    @pytest.mark.asyncio
    async def test_empty_list(self, fake_llm_provider: ProviderTest) -> None:
        assert await _filter_out_users_with_wrong_keys(fake_llm_provider, _llm_config, []) == []

    @pytest.mark.asyncio
    async def test_all_working(self, fake_llm_provider: ProviderTest, five_user_key_infos: list[UserKeyInfo]) -> None:
        assert five_user_key_infos[1].api_key
        assert five_user_key_infos[3].api_key

        fake_llm_provider.api_keys_statuses.set(five_user_key_infos[1].api_key, KeyStatus.broken)
        fake_llm_provider.api_keys_statuses.set(five_user_key_infos[3].api_key, KeyStatus.quota)

        infos = await _filter_out_users_with_wrong_keys(fake_llm_provider, _llm_config, five_user_key_infos)

        assert infos == [five_user_key_infos[i] for i in [0, 2, 4]]


class TestFilterOutUsersWithoutKeys:
    @pytest.mark.asyncio
    async def test_empty_list(self) -> None:
        assert await _filter_out_users_without_keys([]) == []

    @pytest.mark.asyncio
    async def test_all_working(self, five_user_key_infos: list[UserKeyInfo]) -> None:
        five_user_key_infos[1] = five_user_key_infos[1].replace(api_key=None)
        five_user_key_infos[3] = five_user_key_infos[3].replace(api_key=None)

        infos = await _filter_out_users_without_keys(five_user_key_infos)

        assert infos == [five_user_key_infos[i] for i in [0, 2, 4]]


class TestFilterOutUsersForWhomeEntryIsTooOld:
    @pytest.mark.asyncio
    async def test_empty_list(self) -> None:
        assert await _filter_out_users_for_whome_entry_is_too_old([], datetime.timedelta(days=1)) == []

    @pytest.mark.asyncio
    async def test_all_working(self, five_user_key_infos: list[UserKeyInfo]) -> None:
        for i, days in enumerate([5, 2, 3, 1, 4]):
            five_user_key_infos[i] = five_user_key_infos[i].replace(
                process_entries_not_older_than=datetime.timedelta(days=days)
            )

        infos = await _filter_out_users_for_whome_entry_is_too_old(five_user_key_infos, datetime.timedelta(days=3))

        assert infos == [five_user_key_infos[i] for i in [0, 2, 4]]


class TestFilterOutUsersWithOverusedKeys:
    @pytest.mark.asyncio
    async def test_empty_list(self) -> None:
        assert await _filter_out_users_with_overused_keys([], reserved_cost=USDCost(Decimal(100))) == []

    @pytest.mark.asyncio
    async def test_all_working(self, five_user_key_infos: list[UserKeyInfo]) -> None:
        for i, max_tokens_cost_in_month in enumerate([201, 100, 300, 200, 500]):
            five_user_key_infos[i] = five_user_key_infos[i].replace(
                cost_used=USDCost(Decimal(50)), max_tokens_cost_in_month=USDCost(Decimal(max_tokens_cost_in_month))
            )

        infos = await _filter_out_users_with_overused_keys(five_user_key_infos, reserved_cost=USDCost(Decimal(150)))

        assert infos == [five_user_key_infos[i] for i in [0, 2, 4]]


class TestChooseUser:
    @pytest.mark.asyncio
    async def test_no_users(self) -> None:
        interval_started_at = month_interval_start()

        assert (
            await _choose_user(infos=[], reserved_cost=USDCost(Decimal(0)), interval_started_at=interval_started_at)
            is None
        )

    @pytest.mark.asyncio
    async def test_no_users_with_resources(self, five_user_key_infos: list[UserKeyInfo]) -> None:
        interval_started_at = month_interval_start()

        reserved_cost = USDCost(max(info.max_tokens_cost_in_month for info in five_user_key_infos) + 1)

        info = await _choose_user(
            infos=five_user_key_infos,
            reserved_cost=reserved_cost,
            interval_started_at=interval_started_at,
        )

        assert info is None

    @pytest.mark.asyncio
    async def test_all_working(self, five_user_key_infos: list[UserKeyInfo]) -> None:
        interval_started_at = month_interval_start()

        max_cost = USDCost(max(info.max_tokens_cost_in_month for info in five_user_key_infos) + 1)

        five_user_key_infos[2] = five_user_key_infos[2].replace(
            max_tokens_cost_in_month=max_cost + five_user_key_infos[2].cost_used
        )

        info = await _choose_user(
            infos=five_user_key_infos, reserved_cost=max_cost, interval_started_at=interval_started_at
        )

        assert info == five_user_key_infos[2]


class TestGetUserKeyInfos:
    @pytest.mark.asyncio
    async def test_no_users(self) -> None:
        interval_started_at = month_interval_start()

        assert await _get_user_key_infos(LLMProvider.test, user_ids=[], interval_started_at=interval_started_at) == []

    @pytest.mark.asyncio
    async def test_works(self, five_internal_user_ids: list[UserId]) -> None:
        from ffun.application.resources import Resource as AppResource
        from ffun.application.user_settings import UserSetting

        interval_started_at = month_interval_start()

        keys = [LLMApiKey(uuid.uuid4().hex) for _ in range(5)]
        max_tokens_cost_in_month = [USDCost(Decimal(i + 1) * 1000) for i in range(5)]
        days = list(range(5))
        used_costs = [USDCost(Decimal(i * 100)) for i in range(5)]
        reserved_costs = [USDCost(Decimal(i * 10)) for i in range(5)]

        for i, user_id in enumerate(five_internal_user_ids):
            await us_domain.save_setting(user_id=user_id, kind=UserSetting.openai_api_key, value=keys[i])

            await us_domain.save_setting(
                user_id=user_id, kind=UserSetting.max_tokens_cost_in_month, value=max_tokens_cost_in_month[i]
            )

            await us_domain.save_setting(
                user_id=user_id, kind=UserSetting.process_entries_not_older_than, value=days[i]
            )

            await r_domain.try_to_reserve(
                user_id=user_id,
                kind=AppResource.tokens_cost,
                interval_started_at=interval_started_at,
                amount=_cost_points.to_points(reserved_costs[i]),
                limit=_cost_points.to_points(max_tokens_cost_in_month[i]),
            )

            await r_domain.convert_reserved_to_used(
                user_id=user_id,
                kind=AppResource.tokens_cost,
                interval_started_at=interval_started_at,
                used=_cost_points.to_points(used_costs[i]),
                reserved=0,
            )

        infos = await _get_user_key_infos(LLMProvider.openai, five_internal_user_ids, interval_started_at)

        assert infos == [
            UserKeyInfo(
                user_id=five_internal_user_ids[i],
                api_key=keys[i],
                max_tokens_cost_in_month=max_tokens_cost_in_month[i],
                process_entries_not_older_than=datetime.timedelta(days=days[i]),
                cost_used=USDCost(used_costs[i] + reserved_costs[i]),
            )
            for i in range(5)
        ]


@pytest.mark.asyncio
async def test_default_filters() -> None:
    assert _filters == (
        _filter_out_users_without_keys,
        _filter_out_users_for_whome_entry_is_too_old,
        _filter_out_users_with_wrong_keys,
        _filter_out_users_with_overused_keys,
    )


class TestGetCandidates:
    @pytest.mark.asyncio
    async def test_no_users(self, fake_llm_provider: ProviderTest, saved_feed_id: FeedId) -> None:
        interval_started_at = month_interval_start()

        assert (
            await _get_candidates(
                llm=fake_llm_provider,
                llm_config=_llm_config,
                feed_ids={saved_feed_id},
                interval_started_at=interval_started_at,
                entry_age=datetime.timedelta(days=1),
                reserved_cost=USDCost(Decimal(100)),
            )
            == []
        )

    @pytest.mark.asyncio
    async def test_filters_used(
        self,
        fake_llm_provider: ProviderTest,
        saved_feed_id: FeedId,
        another_saved_feed_id: FeedId,
        five_internal_user_ids: list[UserId],
    ) -> None:
        for user_id in five_internal_user_ids[:2]:
            await fl_domain.add_link(user_id, saved_feed_id)

        for user_id in five_internal_user_ids[2:]:
            await fl_domain.add_link(user_id, another_saved_feed_id)

        interval_started_at = month_interval_start()

        filter_1_users = [five_internal_user_ids[0], five_internal_user_ids[2], five_internal_user_ids[4]]
        filter_2_users = five_internal_user_ids
        filter_3_users = five_internal_user_ids[1:]

        def create_filter(filter_users: list[UserId]) -> Any:
            async def _filter(infos: list[UserKeyInfo], **kwargs: Any) -> list[UserKeyInfo]:
                return [info for info in infos if info.user_id in filter_users]

            return _filter

        infos = await _get_candidates(
            llm=fake_llm_provider,
            llm_config=_llm_config,
            feed_ids={saved_feed_id, another_saved_feed_id},
            interval_started_at=interval_started_at,
            entry_age=datetime.timedelta(days=1),
            reserved_cost=USDCost(Decimal(100)),
            filters=(create_filter(filter_1_users), create_filter(filter_2_users), create_filter(filter_3_users)),
        )

        assert {info.user_id for info in infos} == {five_internal_user_ids[2], five_internal_user_ids[4]}

    @pytest.mark.asyncio
    async def test_all_users_excluded(
        self, fake_llm_provider: ProviderTest, saved_feed_id: FeedId, five_internal_user_ids: list[UserId]
    ) -> None:
        for user_id in five_internal_user_ids:
            await fl_domain.add_link(user_id, saved_feed_id)

        interval_started_at = month_interval_start()

        filter_1_users = [five_internal_user_ids[0], five_internal_user_ids[2], five_internal_user_ids[4]]

        def create_filter(filter_users: list[UserId]) -> Any:
            async def _filter(infos: list[UserKeyInfo], **kwargs: Any) -> list[UserKeyInfo]:
                return [info for info in infos if info.user_id in filter_users]

            return _filter

        async def _filter_3(infos: list[UserKeyInfo], **kwargs: Any) -> list[UserKeyInfo]:
            raise Exception("Should not be called")

        infos = await _get_candidates(
            llm=fake_llm_provider,
            llm_config=_llm_config,
            feed_ids={saved_feed_id},
            interval_started_at=interval_started_at,
            entry_age=datetime.timedelta(days=1),
            reserved_cost=USDCost(Decimal(100)),
            filters=(create_filter(filter_1_users), create_filter([]), _filter_3),
        )

        assert infos == []


class TestFindBestUserWithKey:
    @pytest.mark.asyncio
    async def test_no_users(self, fake_llm_provider: ProviderTest, saved_feed_id: FeedId) -> None:
        info = await _find_best_user_with_key(
            llm=fake_llm_provider,
            llm_config=_llm_config,
            feed_ids={saved_feed_id},
            entry_age=datetime.timedelta(days=1),
            interval_started_at=month_interval_start(),
            reserved_cost=USDCost(Decimal(100)),
        )

        assert info is None

    @pytest.mark.asyncio
    async def test_works(
        self, fake_llm_provider: ProviderTest, saved_feed_id: FeedId, five_user_key_infos: list[UserKeyInfo]
    ) -> None:
        for user_key_info in five_user_key_infos:
            await fl_domain.add_link(user_key_info.user_id, saved_feed_id)

        chosen_users: set[UserId] = set()

        interval_started_at = month_interval_start()

        used_cost = USDCost(Decimal(7))

        for _ in range(len(five_user_key_infos)):
            info = await _find_best_user_with_key(
                llm=fake_llm_provider,
                llm_config=_llm_config,
                feed_ids={saved_feed_id},
                entry_age=datetime.timedelta(days=0),
                interval_started_at=interval_started_at,
                reserved_cost=used_cost,
            )

            assert info is not None
            assert info.cost_used == five_user_key_infos[0].cost_used

            chosen_users.add(info.user_id)

            await r_domain.convert_reserved_to_used(
                user_id=info.user_id,
                kind=AppResource.tokens_cost,
                interval_started_at=interval_started_at,
                used=_cost_points.to_points(used_cost),
                reserved=_cost_points.to_points(used_cost),
            )

        assert chosen_users == {info.user_id for info in five_user_key_infos}

        # next user will have more used tokens
        info = await _find_best_user_with_key(
            llm=fake_llm_provider,
            llm_config=_llm_config,
            feed_ids={saved_feed_id},
            entry_age=datetime.timedelta(days=0),
            interval_started_at=interval_started_at,
            reserved_cost=used_cost,
        )

        assert info is not None
        assert info.cost_used == USDCost(five_user_key_infos[0].cost_used + used_cost)


@pytest.fixture
def select_key_context(saved_feed_id: FeedId) -> SelectKeyContext:
    return SelectKeyContext(
        llm_config=_llm_config,
        collections_api_key=None,
        general_api_key=None,
        feed_ids={saved_feed_id},
        entry_age=datetime.timedelta(seconds=0),
        reserved_cost=USDCost(Decimal(0)),
    )


class TestChooseGeneralKey:

    @pytest.mark.asyncio
    async def test_no_general_key_specified(
        self, fake_llm_provider: ProviderTest, select_key_context: SelectKeyContext
    ) -> None:
        assert select_key_context.general_api_key is None

        assert await _choose_general_key(fake_llm_provider, select_key_context) is None

    @pytest.mark.asyncio
    async def test_general_key_specified(
        self, fake_llm_provider: ProviderTest, select_key_context: SelectKeyContext, fake_llm_api_key: LLMApiKey
    ) -> None:

        select_key_context = select_key_context.replace(general_api_key=LLMGeneralApiKey(fake_llm_api_key))

        usage = await _choose_general_key(fake_llm_provider, select_key_context)

        assert usage == APIKeyUsage(
            provider=fake_llm_provider.provider,
            user_id=None,
            api_key=fake_llm_api_key,
            reserved_cost=select_key_context.reserved_cost,
            used_cost=None,
            input_tokens=None,
            output_tokens=None,
            interval_started_at=select_key_context.interval_started_at,
        )


class TestChooseCollectionsKey:

    @pytest.mark.asyncio
    async def test_no_collections_key_specified(
        self, fake_llm_provider: ProviderTest, select_key_context: SelectKeyContext, another_saved_feed_id: FeedId
    ) -> None:
        select_key_context.feed_ids.add(another_saved_feed_id)

        assert select_key_context.collections_api_key is None
        assert len(select_key_context.feed_ids) > 1

        assert await _choose_collections_key(fake_llm_provider, select_key_context) is None

    @pytest.mark.parametrize("collection_feed_index", [0, 1])
    @pytest.mark.asyncio
    async def test_collections_key_specified__in_collection(  # noqa: ignore=CFQ002
        self,
        fake_llm_provider: ProviderTest,
        select_key_context: SelectKeyContext,
        fake_llm_api_key: LLMApiKey,
        collection_id_for_test_feeds: CollectionId,
        another_saved_feed_id: FeedId,
        collection_feed_index: int,
    ) -> None:
        select_key_context = select_key_context.replace(
            collections_api_key=LLMCollectionApiKey(fake_llm_api_key),
            feed_ids=select_key_context.feed_ids.union({another_saved_feed_id}),
        )

        await collections.add_test_feed_to_collections(
            collection_id_for_test_feeds, list(select_key_context.feed_ids)[collection_feed_index]
        )

        usage = await _choose_collections_key(fake_llm_provider, select_key_context)

        assert usage == APIKeyUsage(
            provider=fake_llm_provider.provider,
            user_id=None,
            api_key=fake_llm_api_key,
            reserved_cost=select_key_context.reserved_cost,
            used_cost=None,
            input_tokens=None,
            output_tokens=None,
            interval_started_at=select_key_context.interval_started_at,
        )

    @pytest.mark.asyncio
    async def test_collections_key_specified__not_in_collection(
        self,
        fake_llm_provider: ProviderTest,
        select_key_context: SelectKeyContext,
    ) -> None:
        key = uuid.uuid4().hex

        select_key_context = select_key_context.replace(collections_api_key=key)

        assert all(not collections.has_feed(feed_id) for feed_id in select_key_context.feed_ids)

        assert await _choose_collections_key(fake_llm_provider, select_key_context) is None


class TestChooseUserKey:

    @pytest.mark.asyncio
    async def test_no_users(self, fake_llm_provider: ProviderTest, select_key_context: SelectKeyContext) -> None:
        assert await _choose_user_key(fake_llm_provider, select_key_context) is None

    @pytest.mark.parametrize("collection_feed_index", [0, 1])
    @pytest.mark.asyncio
    async def test_protection_from_collections_processing(
        self,
        fake_llm_provider: ProviderTest,
        select_key_context: SelectKeyContext,
        collection_id_for_test_feeds: CollectionId,
        another_saved_feed_id: FeedId,
        collection_feed_index: int,
    ) -> None:
        select_key_context.feed_ids.add(another_saved_feed_id)
        await collections.add_test_feed_to_collections(
            collection_id_for_test_feeds, list(select_key_context.feed_ids)[collection_feed_index]
        )

        with pytest.raises(errors.FeedsFromCollectionsMustNotBeProcessedWithUserAPIKeys):
            await _choose_user_key(fake_llm_provider, select_key_context)

    @pytest.mark.asyncio
    async def test_found_user(
        self,
        fake_llm_provider: ProviderTest,
        select_key_context: SelectKeyContext,
        saved_feed_id: FeedId,
        five_user_key_infos: list[UserKeyInfo],
    ) -> None:
        info = five_user_key_infos[0]

        assert info.api_key is not None

        await fl_domain.add_link(info.user_id, saved_feed_id)

        usage = await _choose_user_key(fake_llm_provider, select_key_context)

        assert usage == APIKeyUsage(
            provider=fake_llm_provider.provider,
            user_id=info.user_id,
            api_key=info.api_key,
            reserved_cost=select_key_context.reserved_cost,
            used_cost=None,
            input_tokens=None,
            output_tokens=None,
            interval_started_at=select_key_context.interval_started_at,
        )


class TestKeySelectors:

    def test_order(self) -> None:
        """Just double protection from overriding this constant"""
        assert _key_selectors == (_choose_collections_key, _choose_general_key, _choose_user_key)


class TestChooseApiKey:

    def create_selector(self, api_key: LLMApiKey | None) -> Any:
        async def success_selector(llm: ProviderInterface, context: SelectKeyContext) -> APIKeyUsage:
            assert api_key is not None

            return APIKeyUsage(
                provider=llm.provider,
                user_id=new_user_id(),
                api_key=api_key,
                reserved_cost=context.reserved_cost,
                used_cost=None,
                input_tokens=None,
                output_tokens=None,
                interval_started_at=context.interval_started_at,
            )

        async def fail_selector(llm: ProviderInterface, context: SelectKeyContext) -> None:
            return None

        return success_selector if api_key is not None else fail_selector

    @pytest.mark.asyncio
    async def test_no_selectors(self, fake_llm_provider: ProviderTest, select_key_context: SelectKeyContext) -> None:
        key_usage = await choose_api_key(fake_llm_provider, select_key_context, selectors=[])
        assert key_usage is None

    @pytest.mark.asyncio
    async def test_key_not_found(self, fake_llm_provider: ProviderTest, select_key_context: SelectKeyContext) -> None:
        selectors = [self.create_selector(None), self.create_selector(None), self.create_selector(None)]

        key_usage = await choose_api_key(fake_llm_provider, select_key_context, selectors=selectors)

        assert key_usage is None

    @pytest.mark.asyncio
    async def test_choose_first(self, fake_llm_provider: ProviderTest, select_key_context: SelectKeyContext) -> None:
        expected_key = LLMApiKey(uuid.uuid4().hex)

        selectors = [
            self.create_selector(None),
            self.create_selector(expected_key),
            self.create_selector(None),
            self.create_selector(LLMApiKey(uuid.uuid4().hex)),
            self.create_selector(None),
        ]

        usage = await choose_api_key(fake_llm_provider, select_key_context, selectors=selectors)

        assert usage is not None

        assert usage.api_key == expected_key


class TestUseApiKey:

    @pytest.mark.asyncio
    async def test_success(self, internal_user_id: UserId, fake_llm_api_key: LLMApiKey) -> None:

        interval_started_at = month_interval_start()

        reserved_cost = USDCost(Decimal(567))

        await r_domain.try_to_reserve(
            user_id=internal_user_id,
            kind=AppResource.tokens_cost,
            interval_started_at=interval_started_at,
            amount=_cost_points.to_points(reserved_cost),
            limit=_cost_points.to_points(USDCost(Decimal(1000))),
        )

        resources = await r_domain.load_resources(
            user_ids=[internal_user_id], kind=AppResource.tokens_cost, interval_started_at=interval_started_at
        )

        used_cost = USDCost(Decimal(132))

        key_usage = APIKeyUsage(
            provider=LLMProvider.test,
            user_id=internal_user_id,
            api_key=fake_llm_api_key,
            reserved_cost=reserved_cost,
            used_cost=None,
            input_tokens=None,
            output_tokens=None,
            interval_started_at=interval_started_at,
        )

        with capture_logs() as logs:
            async with use_api_key(key_usage):
                key_usage.used_cost = used_cost

        assert_logs_has_business_event(
            logs,
            "llm_api_key_used",
            user_id=internal_user_id,
            llm_provider=LLMProvider.test,
            new_key_status=KeyStatus.unknown,
        )

        resources = await r_domain.load_resources(
            user_ids=[internal_user_id], kind=AppResource.tokens_cost, interval_started_at=interval_started_at
        )

        assert resources == {
            internal_user_id: r_entities.Resource(
                user_id=internal_user_id,
                kind=AppResource.tokens_cost,
                interval_started_at=interval_started_at,
                used=_cost_points.to_points(used_cost),
                reserved=0,
            )
        }

    @pytest.mark.asyncio
    async def test_no_used_tokens_specified(self, internal_user_id: UserId, fake_llm_api_key: LLMApiKey) -> None:

        interval_started_at = month_interval_start()

        reserved_cost = USDCost(Decimal(567))

        await r_domain.try_to_reserve(
            user_id=internal_user_id,
            kind=AppResource.tokens_cost,
            interval_started_at=interval_started_at,
            amount=_cost_points.to_points(reserved_cost),
            limit=_cost_points.to_points(USDCost(Decimal(1000))),
        )

        resources = await r_domain.load_resources(
            user_ids=[internal_user_id], kind=AppResource.tokens_cost, interval_started_at=interval_started_at
        )

        key_usage = APIKeyUsage(
            provider=LLMProvider.test,
            user_id=internal_user_id,
            api_key=fake_llm_api_key,
            reserved_cost=reserved_cost,
            used_cost=None,
            input_tokens=None,
            output_tokens=None,
            interval_started_at=interval_started_at,
        )

        with capture_logs() as logs:
            with pytest.raises(errors.UsedTokensHasNotSpecified):
                async with use_api_key(key_usage):
                    assert key_usage.used_cost is None
                    llm_providers.get(key_usage.provider).provider.api_keys_statuses.set(
                        key_usage.api_key, KeyStatus.works
                    )

        assert_logs_has_business_event(
            logs,
            "llm_api_key_used",
            user_id=internal_user_id,
            llm_provider=LLMProvider.test,
            new_key_status=KeyStatus.works,
        )

        resources = await r_domain.load_resources(
            user_ids=[internal_user_id], kind=AppResource.tokens_cost, interval_started_at=interval_started_at
        )

        assert resources == {
            internal_user_id: r_entities.Resource(
                user_id=internal_user_id,
                kind=AppResource.tokens_cost,
                interval_started_at=interval_started_at,
                used=_cost_points.to_points(reserved_cost),
                reserved=0,
            )
        }

    @pytest.mark.asyncio
    async def test_exception_in_child_code(self, internal_user_id: UserId, fake_llm_api_key: LLMApiKey) -> None:

        interval_started_at = month_interval_start()

        reserved_cost = USDCost(Decimal(567))

        await r_domain.try_to_reserve(
            user_id=internal_user_id,
            # TODO: differentiate tokens in all places
            kind=AppResource.tokens_cost,
            interval_started_at=interval_started_at,
            amount=_cost_points.to_points(reserved_cost),
            limit=_cost_points.to_points(USDCost(Decimal(1000))),
        )

        resources = await r_domain.load_resources(
            user_ids=[internal_user_id], kind=AppResource.tokens_cost, interval_started_at=interval_started_at
        )

        class FakeError(Exception):
            pass

        key_usage = APIKeyUsage(
            provider=LLMProvider.test,
            user_id=internal_user_id,
            api_key=fake_llm_api_key,
            reserved_cost=reserved_cost,
            used_cost=None,
            input_tokens=None,
            output_tokens=None,
            interval_started_at=interval_started_at,
        )

        with capture_logs() as logs:
            with pytest.raises(FakeError):
                async with use_api_key(key_usage):
                    raise FakeError()

        assert_logs_has_business_event(
            logs,
            "llm_api_key_used",
            user_id=internal_user_id,
            llm_provider=LLMProvider.test,
            new_key_status=KeyStatus.unknown,
        )

        resources = await r_domain.load_resources(
            user_ids=[internal_user_id], kind=AppResource.tokens_cost, interval_started_at=interval_started_at
        )

        assert resources == {
            internal_user_id: r_entities.Resource(
                user_id=internal_user_id,
                kind=AppResource.tokens_cost,
                interval_started_at=interval_started_at,
                used=_cost_points.to_points(reserved_cost),
                reserved=0,
            )
        }

    @pytest.mark.asyncio
    async def test_no_user_in_key_usage(self, fake_llm_api_key: LLMApiKey) -> None:

        interval_started_at = month_interval_start()

        reserved_cost = USDCost(Decimal(567))
        used_cost = USDCost(Decimal(214))

        key_usage = APIKeyUsage(
            provider=LLMProvider.test,
            user_id=None,
            api_key=fake_llm_api_key,
            reserved_cost=reserved_cost,
            used_cost=None,
            input_tokens=None,
            output_tokens=None,
            interval_started_at=interval_started_at,
        )

        with capture_logs() as logs:
            async with use_api_key(key_usage):
                key_usage.used_cost = used_cost

        assert_logs_has_business_event(
            logs,
            "llm_api_key_used",
            user_id=None,
            llm_provider=LLMProvider.test,
            new_key_status=KeyStatus.unknown,
        )



from typing import Iterable

from ffun.domain.entities import TagId
from ffun.scores import entities, operations

count_rules_per_user = operations.count_rules_per_user


def get_score_rules(rules: Iterable[entities.Rule], tags: set[TagId]) -> list[entities.Rule]:
    score_rules = []

    for rule in rules:
        if rule.excluded_tags & tags:
            continue

        if rule.required_tags <= tags:
            score_rules.append(rule)

    return score_rules


def get_score_contributions(rules: Iterable[entities.Rule], tags: set[TagId]) -> tuple[int, dict[TagId, int]]:
    score = 0
    contributions: dict[TagId, int] = {}

    for rule in get_score_rules(rules, tags):
        score += rule.score

        # We may want to think about calculating contributions for excluded tags
        # but currently there is no visible need for that
        for tag in rule.required_tags:
            contributions[tag] = contributions.get(tag, 0) + rule.score

    return score, contributions


create_or_update_rule = operations.create_or_update_rule
delete_rule = operations.delete_rule
get_rules = operations.get_rules
update_rule = operations.update_rule



from ffun.core import errors


class Error(errors.Error):
    pass


class ClientError(Error):
    pass


class AuthError(ClientError):
    pass


class QuotaError(ClientError):
    pass


class UnknownError(ClientError):
    pass


class ProhibitedContentError(ClientError):
    pass


class MalformedFunctionCallError(ClientError):
    pass


class UnknownFinishReasonError(ClientError):
    pass


class PromptBlocked(ClientError):
    pass



import pytest
from pytest_mock import MockerFixture

from ffun.auth.domain import (
    logout_user_from_all_sessions,
    logout_user_from_all_sessions_in_service,
    remove_user_from_external_service,
)
from ffun.domain.entities import UserId
from ffun.users.entities import Service


class TestRemoveUserFromExternalService:

    @pytest.mark.asyncio
    async def test_remove_supertokens_user(self, external_user_id: str, mocker: MockerFixture) -> None:
        remove_supertokens_user = mocker.patch("ffun.auth.supertokens.remove_user")

        await remove_user_from_external_service(Service.supertokens, external_user_id)

        remove_supertokens_user.assert_called_once_with(external_user_id)

    @pytest.mark.asyncio
    async def test_remove_single_user(self, external_user_id: str, mocker: MockerFixture) -> None:
        remove_supertokens_user = mocker.patch("ffun.auth.supertokens.remove_user")

        await remove_user_from_external_service(Service.single, external_user_id)

        remove_supertokens_user.assert_not_called()


class TestLogoutUserFromAllSessionsInService:

    @pytest.mark.asyncio
    async def test_logout_supertokens_user(self, external_user_id: str, mocker: MockerFixture) -> None:
        logout_supertokens_user = mocker.patch("ffun.auth.supertokens.logout_user_from_all_sessions")

        await logout_user_from_all_sessions_in_service(Service.supertokens, external_user_id)

        logout_supertokens_user.assert_called_once_with(external_user_id)

    @pytest.mark.asyncio
    async def test_logout_single_user(self, external_user_id: str, mocker: MockerFixture) -> None:
        logout_supertokens_user = mocker.patch("ffun.auth.supertokens.logout_user_from_all_sessions")

        await logout_user_from_all_sessions_in_service(Service.single, external_user_id)

        logout_supertokens_user.assert_not_called()


class TestLogoutUserFromAllSessions:

    @pytest.mark.asyncio
    async def test_logout_user(self, internal_user_id: UserId, mocker: MockerFixture) -> None:
        get_user_external_ids = mocker.patch("ffun.users.domain.get_user_external_ids")
        get_user_external_ids.return_value = {Service.supertokens: "supertokens_id", Service.single: "single_id"}

        logout_user = mocker.patch("ffun.auth.domain.logout_user_from_all_sessions_in_service")

        await logout_user_from_all_sessions(internal_user_id)

        assert logout_user.call_count == 2

        logout_user.assert_any_call(Service.supertokens, "supertokens_id")
        logout_user.assert_any_call(Service.single, "single_id")



import contextlib
from typing import Generator, Sequence

from ffun.core import logging
from ffun.google import errors
from ffun.google.client import Client
from ffun.google.entities import ChatMessage, GenerationConfig, GoogleChatRequest, GoogleChatResponse
from ffun.llms_framework import domain as llmsf_domain
from ffun.llms_framework import errors as llmsf_errors
from ffun.llms_framework.entities import KeyStatus, LLMConfiguration, LLMProvider
from ffun.llms_framework.keys_statuses import Statuses
from ffun.llms_framework.provider_interface import ProviderInterface

logger = logging.get_module_logger()

# TODO: allow specifying providers as third-party plugins
#       or create an issue on github about that

# TODO: Solve conflict with using different providers by users vs lack of using providers by system logic (collections)
#       There is possible situation when collections uses one provider and ignore another,
#       but user wants to use another provider
#       This may lead to processing all feeds in collections with the user key
#       Maybe it is better to restrict processing of collections with user keys
#       Maybe, we could move flags: for_collections, for_users, for_all to the configuration of all tag processors?


@contextlib.contextmanager
def track_key_status(key: str, statuses: Statuses) -> Generator[None, None, None]:
    try:

        yield

        statuses.set(key, KeyStatus.works)

    except errors.AuthError:
        statuses.set(key, KeyStatus.broken)
        raise

    except errors.QuotaError:
        statuses.set(key, KeyStatus.quota)
        raise

    except Exception:
        statuses.set(key, KeyStatus.unknown)
        raise


# TODO: add to the documentation note about quotas of Gemini API (especially free tier)
# TODO: tests
class GoogleInterface(ProviderInterface):
    provider = LLMProvider.google

    def estimate_tokens(self, config: LLMConfiguration, text: str) -> int:
        # There are multiple ways to count tokens for Gemini:
        # 1. Use vertextai lib to count tokens locally.
        #    For now, it is too big a dependency for the project.
        #    Also, it was difficult to integrate official Google Libs.
        # 2. Use Google REST API to calculate tokens remotely.
        #    Looks like too much overhead.
        # 3. Estimate upper boundaries with a heuristic.
        #    Because the context windows of Gemini models are big
        #    this approach looks ok for start.

        # Coefficient was chosen according to ChatGPT's recommendations :-)
        return int(len(text) * 1.8)

    async def chat_request(  # type: ignore
        self, config: LLMConfiguration, api_key: str, request: GoogleChatRequest
    ) -> GoogleChatResponse:

        client = Client(api_key=api_key)

        generation_config = GenerationConfig(
            max_output_tokens=config.max_return_tokens,
            temperature=float(config.temperature),
            top_p=float(config.top_p),
            top_k=None,
        )

        try:
            with track_key_status(api_key, self.api_keys_statuses):
                answer = await client.generate_content(model=config.model, request=request, config=generation_config)

        except errors.ClientError as e:
            message = str(e)
            logger.info("google_api_error", message=message)
            raise llmsf_errors.TemporaryError(message=message) from e

        logger.info("google_response")

        return answer

    def prepare_requests(self, config: LLMConfiguration, text: str) -> Sequence[GoogleChatRequest]:
        parts = llmsf_domain.split_text_according_to_tokens(llm=self, llm_config=config, text=text)

        requests = []

        for part in parts:
            request = GoogleChatRequest(
                system=ChatMessage(text=config.system), messages=[ChatMessage(role="user", text=part)]
            )

            requests.append(request)

        return requests

    async def check_api_key(self, config: LLMConfiguration, api_key: str) -> KeyStatus:
        client = Client(api_key=api_key)

        try:
            with track_key_status(api_key, self.api_keys_statuses):
                await client.list_models()
        except errors.ClientError:
            pass

        return self.api_keys_statuses.get(api_key)


provider = GoogleInterface()



import datetime
import uuid

import pydantic


class Resource(pydantic.BaseModel):
    user_id: uuid.UUID
    kind: int
    interval_started_at: datetime.datetime

    used: int
    reserved: int

    @property
    def total(self) -> int:
        return self.used + self.reserved



import pytest

from ffun.librarian import errors
from ffun.librarian.processors.llm_general import Processor
from ffun.librarian.tag_extractors import dog_tags_extractor
from ffun.librarian.text_cleaners import clear_nothing
from ffun.library.entities import Entry
from ffun.llms_framework.entities import LLMApiKey, LLMConfiguration, LLMGeneralApiKey, LLMProvider, LLMTokens
from ffun.llms_framework.provider_interface import ChatResponseTest
from ffun.ontology.entities import ProcessorTag


class TestProcessor:

    @pytest.fixture
    def llm_config(self) -> LLMConfiguration:
        return LLMConfiguration(
            model="test-model-1",
            system="system prompt",
            max_return_tokens=LLMTokens(143),
            text_parts_intersection=100,
            temperature=0,
            top_p=0,
            presence_penalty=0,
            frequency_penalty=0,
        )

    @pytest.fixture
    def llm_processor(self, llm_config: LLMConfiguration) -> Processor:
        return Processor(
            name="test-llm-processor",
            llm_provider=LLMProvider.test,
            llm_config=llm_config,
            entry_template="{entry.title} {entry.body}",
            text_cleaner=clear_nothing,
            tag_extractor=dog_tags_extractor,
            collections_api_key=None,
            general_api_key=None,
        )

    def test_extract_tags(self, llm_processor: Processor) -> None:

        responses = [
            ChatResponseTest(
                content="@tag-1 @tag-2",
            ),
            ChatResponseTest(content="@tag-3 @tag-2"),
        ]

        tags = llm_processor.extract_tags(responses)

        tags.sort(key=lambda x: x.raw_uid)

        assert tags == [ProcessorTag(raw_uid="tag-1"), ProcessorTag(raw_uid="tag-2"), ProcessorTag(raw_uid="tag-3")]

    @pytest.mark.asyncio
    async def test_process__no_api_key_found(self, llm_processor: Processor, cataloged_entry: Entry) -> None:
        with pytest.raises(errors.SkipEntryProcessing):
            await llm_processor.process(cataloged_entry)

    @pytest.mark.asyncio
    async def test_process__has_api_key_found(
        self, llm_processor: Processor, cataloged_entry: Entry, fake_llm_api_key: LLMApiKey
    ) -> None:

        entry = cataloged_entry.replace(title="@tag-1 @tag-2", body="@tag-3 @tag-2")

        llm_processor.general_api_key = LLMGeneralApiKey(fake_llm_api_key)

        tags = await llm_processor.process(entry)

        tags.sort(key=lambda x: x.raw_uid)

        assert tags == [ProcessorTag(raw_uid="tag-1"), ProcessorTag(raw_uid="tag-2"), ProcessorTag(raw_uid="tag-3")]

    @pytest.mark.asyncio
    async def test_process__temporary_error_processing(
        self, llm_processor: Processor, cataloged_entry: Entry, fake_llm_api_key: LLMApiKey
    ) -> None:

        entry = cataloged_entry.replace(title="@tag-1 @tag-2", body="raise TemporaryError")

        llm_processor.general_api_key = LLMGeneralApiKey(fake_llm_api_key)

        with pytest.raises(errors.TemporaryErrorInProcessor):
            await llm_processor.process(entry)



"""
add-feeds-links
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__: set[str] = set()

sql_create_feeds_links_table = """
CREATE TABLE fl_links (
    id UUID PRIMARY KEY,
    feed_id UUID NOT NULL,
    user_id UUID NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW())
"""

sql_create_unique_index = """
CREATE UNIQUE INDEX idx_fl_links_user_id_feed_id ON fl_links (user_id, feed_id)
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_feeds_links_table)
    cursor.execute(sql_create_unique_index)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE fl_links")


steps = [step(apply_step, rollback_step)]



from ffun.core import errors


class Error(errors.Error):
    pass


class DuplicateCollectionIds(Error):
    pass


class DuplicateCollectionSlugs(Error):
    pass


class CollectionNotFound(Error):
    pass


class DuplicateCollectionOrders(Error):
    pass


class CollectionIsEmpty(Error):
    pass


class FeedNotFound(Error):
    pass



import contextlib
import copy
import datetime
import enum
import functools
import inspect
import logging
import sys
import time
import uuid
from collections import abc
from typing import Any, Callable, ContextManager, Iterable, Iterator, Protocol, TypeVar

import pydantic_settings
import structlog
from sentry_sdk import capture_exception, capture_message
from structlog import contextvars as structlog_contextvars

from ffun.core import errors
from ffun.domain.entities import UserId

LabelValue = int | str | None


class Renderer(str, enum.Enum):
    console = "console"
    json = "json"


class Settings(pydantic_settings.BaseSettings):
    level: str = "INFO"
    renderer: Renderer = Renderer.console

    model_config = pydantic_settings.SettingsConfigDict(
        env_nested_delimiter="__", env_file=".env", env_prefix="FFUN_LOGS_", extra="allow"
    )

    @property
    def structlog_level(self) -> int:
        return getattr(logging, self.level.upper())  # type: ignore


settings = Settings()


class LogProcessorType(Protocol):
    async def __call__(self, _: Any, __: Any, event_dict: dict[str, Any]) -> dict[str, Any]:
        pass


class Formatter:
    __slots__ = ()

    def can_format(self, value: Any) -> bool:
        raise NotImplementedError('You must implement "can_format" in child class')

    def format(self, value: Any) -> Any:
        raise NotImplementedError('You must implement "format" in child class')


class DateFormatter(Formatter):
    __slots__ = ()

    def can_format(self, value: Any) -> bool:
        return isinstance(value, (datetime.date, datetime.datetime))

    def format(self, value: Any) -> Any:
        return value.isoformat()


class UUIDFormatter(Formatter):
    __slots__ = ()

    def can_format(self, value: Any) -> bool:
        return isinstance(value, uuid.UUID)

    def format(self, value: Any) -> Any:
        return str(value)


class EnumFormatter(Formatter):
    __slots__ = ()

    def can_format(self, value: Any) -> bool:
        return isinstance(value, enum.Enum)

    def format(self, value: Any) -> Any:
        return value.name


class ProcessorFormatter:
    __slots__ = ("_formatters",)

    def __init__(self, formatters: Iterable[Formatter]) -> None:
        self._formatters = list(formatters)

    def __call__(self, _: Any, __: Any, event_dict: dict[str, Any]) -> dict[str, Any]:
        for key, value in event_dict.items():
            for formatter in self._formatters:
                if formatter.can_format(value):
                    event_dict[key] = formatter.format(value)
                    break

        return event_dict


def info_extracter(_: Any, __: Any, event_dict: dict[str, Any]) -> dict[str, Any]:
    replaced = False

    for key, value in list(event_dict.items()):
        if not hasattr(value, "log_info"):
            continue

        for k, v in value.log_info().items():
            event_dict[f"{key}_{k}"] = v

        del event_dict[key]

        replaced = True

    if not replaced:
        return event_dict

    return info_extracter(_, __, event_dict)


def create_formatter() -> ProcessorFormatter:
    formatters = [DateFormatter(), UUIDFormatter(), EnumFormatter()]
    return ProcessorFormatter(formatters)


def log_errors_to_sentry(_: Any, __: Any, event_dict: dict[str, Any]) -> dict[str, Any]:
    if event_dict.get("sentry_skip"):
        return event_dict

    if event_dict.get("level", "").upper() != "ERROR":
        return event_dict

    if event_dict.get("exc_info"):
        capture_exception(sys.exc_info())
    else:
        capture_message(event_dict["event"])

    return event_dict


def processors_list(use_sentry: bool) -> list[LogProcessorType]:
    sentry_processor = None

    if use_sentry:
        sentry_processor = log_errors_to_sentry

    processors_list = [
        structlog.contextvars.merge_contextvars,
        info_extracter,
        structlog.processors.add_log_level,
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        sentry_processor,
        structlog.processors.TimeStamper(fmt="ISO", utc=True, key="timestamp"),
        create_formatter(),
        structlog.dev.ConsoleRenderer() if settings.renderer == Renderer.console else None,
        structlog.processors.JSONRenderer() if settings.renderer == Renderer.json else None,
    ]

    return [p for p in processors_list if p is not None]  # type: ignore


class FFunBoundLogger(structlog.typing.FilteringBoundLogger):
    def measure(self, event: str, value: float | int, **labels: LabelValue) -> None:
        pass

    def measure_block_time(  # type: ignore
        self, event: str, **labels: LabelValue
    ) -> ContextManager[dict[str, LabelValue]]:
        pass

    def _normalize_value(self, value: Any) -> Any:
        pass

    def business_event(self, event: str, user_id: UserId | None, **attributes: Any) -> Any:
        pass

    def business_slice(self, event: str, user_id: UserId | None, **attributes: LabelValue) -> Any:
        pass


class MeasuringBoundLoggerMixin:
    """We extend a logger class with additional metrics logic.

    This mixin is required to work with metrics in 100% the same way as with logs.

    By calling `logging.measure(...)` and other methods, we guarantee that metrics will have the exact attributes
    as expected from log messages (both custom and general like "module"), so the third-party software
    would be able to filter/process messages universally.
    """

    def measure(self, event: str, value: float | int, **labels: LabelValue) -> Any:
        if not labels:
            return self.info(event, m_kind="measure", m_value=value)  # type: ignore

        with bound_measure_labels(**labels):
            return self.info(event, m_kind="measure", m_value=value)  # type: ignore

    @contextlib.contextmanager
    def measure_block_time(self, event: str, **labels: LabelValue) -> Iterator[dict[str, LabelValue]]:
        started_at = time.monotonic()

        extra_labels: dict[str, LabelValue] = {}

        with bound_measure_labels(**labels):
            try:
                yield extra_labels
            finally:
                self.measure(event, time.monotonic() - started_at, **extra_labels)


class BusinessBoundLoggerMixin:
    """We extend a logger class with additional business events logic.

    This mixin is required to work with business events in 100% the same way as with logs.
    """

    def _normalize_value(self, value: Any) -> Any:
        if isinstance(value, int | str | float | None):
            return value

        if isinstance(value, abc.Mapping):
            return {str(k): self._normalize_value(v) for k, v in value.items()}

        if isinstance(value, abc.Sequence):
            return [self._normalize_value(v) for v in value]

        return str(value)

    def business_event(self, event: str, user_id: UserId | None, **attributes: LabelValue) -> Any:
        return self.info(  # type: ignore
            event,
            b_kind="event",
            b_user_id=str(user_id) if user_id is not None else None,
            b_uid=str(uuid.uuid4()),
            b_attributes=self._normalize_value(attributes),
        )

    # TODO: test
    def business_slice(self, event: str, user_id: UserId | None, **attributes: LabelValue) -> Any:
        return self.info(  # type: ignore
            event,
            b_kind="slice",
            b_user_id=str(user_id) if user_id is not None else None,
            b_uid=str(uuid.uuid4()),
            b_attributes=self._normalize_value(attributes),
        )


def make_measuring_bound_logger(level: int) -> type[FFunBoundLogger]:
    filtering_logger_class = structlog.make_filtering_bound_logger(level)

    class _FFunBoundLogger(
        MeasuringBoundLoggerMixin, BusinessBoundLoggerMixin, filtering_logger_class  # type: ignore
    ):
        pass

    return _FFunBoundLogger


def initialize(use_sentry: bool) -> None:
    structlog.configure(
        processors=processors_list(use_sentry=use_sentry),  # type: ignore
        wrapper_class=make_measuring_bound_logger(settings.structlog_level),
        context_class=dict,
        logger_factory=structlog.WriteLoggerFactory(),
        cache_logger_on_first_use=True,
    )


def get_module_logger() -> FFunBoundLogger:
    caller_frame = inspect.currentframe().f_back  # type: ignore
    module = inspect.getmodule(caller_frame)
    return structlog.get_logger(module=module.__name__)  # type: ignore


FUNC = TypeVar("FUNC", bound=Callable[..., Any])


class Constructor:
    __slots__ = ("name",)

    def __init__(self, name: str) -> None:
        self.name = name

    def __call__(self, kwargs: dict[str, Any]) -> Any:
        raise NotImplementedError('You must implement "__call__" in child class')


class IdentityConstructor(Constructor):
    __slots__ = ()

    def __init__(self, arg: str) -> None:
        super().__init__(name=arg)

    def __call__(self, kwargs: dict[str, Any]) -> Any:
        return kwargs.get(self.name)


class ArgumentConstructor(Constructor):
    __slots__ = ("key", "attribute")

    def __init__(self, arg: str) -> None:
        super().__init__(name=arg.replace(".", "_"))
        self.key, self.attribute = arg.split(".", 1)

    def __call__(self, kwargs: dict[str, Any]) -> Any:
        if self.key not in kwargs:
            return None

        return getattr(kwargs[self.key], self.attribute, None)


def function_args_to_log(*args: str) -> Callable[[FUNC], FUNC]:
    constructors: list[Constructor] = []

    for arg in args:
        if "." not in arg:
            constructors.append(IdentityConstructor(arg))
        else:
            constructors.append(ArgumentConstructor(arg))

    def wrapper(func: FUNC) -> FUNC:
        @functools.wraps(func)
        def wrapped(**kwargs: Any) -> Any:
            with structlog_contextvars.bound_contextvars(**{c.name: c(kwargs) for c in constructors}):
                return func(**kwargs)

        @functools.wraps(func)
        async def async_wrapped(**kwargs: Any) -> Any:
            with structlog_contextvars.bound_contextvars(**{c.name: c(kwargs) for c in constructors}):
                return await func(**kwargs)

        if inspect.iscoroutinefunction(func):
            return async_wrapped  # type: ignore

        return wrapped  # type: ignore

    return wrapper


@contextlib.contextmanager
def bound_log_args(**kwargs: Any) -> Iterator[None]:

    if not kwargs:
        yield
        return

    if kwargs.keys() & {"m_labels", "m_value", "m_kind", "b_kind"}:
        raise errors.ReservedLogArguments()

    bound_vars = structlog_contextvars.get_contextvars()

    if bound_vars.keys() & kwargs.keys():
        raise errors.DuplicatedLogArguments()

    with structlog_contextvars.bound_contextvars(**kwargs):
        yield


@contextlib.contextmanager
def bound_measure_labels(**labels: LabelValue) -> Iterator[None]:
    if not labels:
        yield
        return

    bound_vars = structlog_contextvars.get_contextvars()

    if "m_labels" in bound_vars:
        if labels.keys() & bound_vars["m_labels"].keys():
            raise errors.DuplicatedMeasureLabels()

        new_labels = copy.copy(bound_vars["m_labels"])

    else:
        new_labels = {}

    new_labels.update(labels)

    with structlog_contextvars.bound_contextvars(m_labels=new_labels):
        yield



import uuid

import httpx
import pytest
from respx.router import MockRouter
from structlog.testing import capture_logs

from ffun.core.tests.helpers import (
    TableSizeDelta,
    TableSizeNotChanged,
    assert_logs,
    assert_logs_have_no_errors,
    assert_logs_levels,
)
from ffun.domain.urls import str_to_absolute_url
from ffun.feeds.entities import FeedError
from ffun.loader import errors
from ffun.loader.entities import ProxyState
from ffun.loader.operations import (
    check_proxy,
    get_proxy_states,
    is_proxy_available,
    load_content,
    update_proxy_states,
)
from ffun.loader.settings import Proxy


class TestUpdateProxyStates:
    @pytest.mark.asyncio
    async def test_nothing_to_update(self) -> None:
        async with TableSizeNotChanged("lr_proxy_states"):
            await update_proxy_states({})

    @pytest.mark.asyncio
    async def test_first_time_updated(self) -> None:
        states = {
            uuid.uuid4().hex: ProxyState.available,
            uuid.uuid4().hex: ProxyState.suspended,
            uuid.uuid4().hex: ProxyState.available,
        }

        async with TableSizeDelta("lr_proxy_states", delta=len(states)):
            await update_proxy_states(states)

        loaded_states = await get_proxy_states(names=list(states.keys()))

        assert loaded_states == states

    @pytest.mark.asyncio
    async def test_second_updated(self) -> None:
        names = [uuid.uuid4().hex for _ in range(4)]

        states = {names[0]: ProxyState.available, names[1]: ProxyState.suspended, names[2]: ProxyState.available}

        await update_proxy_states(states)

        second_states = {
            names[1]: ProxyState.available,
            names[2]: ProxyState.available,
            names[3]: ProxyState.suspended,
        }

        async with TableSizeDelta("lr_proxy_states", delta=1):
            await update_proxy_states(second_states)

        loaded_states = await get_proxy_states(names=names)

        assert loaded_states == {
            names[0]: ProxyState.available,
            names[1]: ProxyState.available,
            names[2]: ProxyState.available,
            names[3]: ProxyState.suspended,
        }


# Most of the functionality are checked in other code
class TestGetProxyStates:
    @pytest.mark.asyncio
    async def test_default_to_available(self) -> None:
        names = [uuid.uuid4().hex for _ in range(4)]

        await update_proxy_states({names[0]: ProxyState.suspended, names[2]: ProxyState.available})

        states = await get_proxy_states(names)

        assert states == {
            names[0]: ProxyState.suspended,
            names[1]: ProxyState.available,
            names[2]: ProxyState.available,
            names[3]: ProxyState.available,
        }


class TestCheckProxy:
    @pytest.mark.asyncio
    async def test_success(self) -> None:
        proxy = Proxy(name=uuid.uuid4().hex, url=None)
        url = "https://www.google.com"
        user_agent = "Mozilla/5.0"

        assert await check_proxy(proxy, url, user_agent)

    @pytest.mark.asyncio
    async def test_error(self) -> None:
        proxy = Proxy(name=uuid.uuid4().hex, url=None)
        url = "localhost:1"
        user_agent = "Mozilla/5.0"

        assert not await check_proxy(proxy, url, user_agent)


class TestIsProxyAvailable:
    @pytest.mark.asyncio
    async def test_success(self) -> None:
        proxy = Proxy(name=uuid.uuid4().hex, url=None)
        anchors = ["https://www.google.com", "https://www.amazon.com"]
        user_agent = "Mozilla/5.0"

        assert await is_proxy_available(proxy, anchors, user_agent)

    @pytest.mark.asyncio
    async def test_error(self) -> None:
        proxy = Proxy(name=uuid.uuid4().hex, url=None)
        anchors = ["localhost:1", "localhost:2"]
        user_agent = "Mozilla/5.0"

        assert not await is_proxy_available(proxy, anchors, user_agent)


class TestLoadContent:

    @pytest.mark.parametrize(
        "bytes_content, expected_headers",
        [
            (b"test-response", {}),
            (
                b"\x1f\x8b\x08\x00v\x18Sf\x02\xff+I-.\xd1-J-.\xc8\xcf+N\x05\x00\xfe\xebMu\r\x00\x00\x00",
                {"Content-Encoding": "gzip"},
            ),
            (b"x\x9c+I-.\xd1-J-.\xc8\xcf+N\x05\x00%A\x05]", {"Content-Encoding": "deflate"}),
            (b"(\xb5/\xfd \ri\x00\x00test-response", {"Content-Encoding": "zstd"}),
            (b"\x1b\x0c\x00\xf8\xa5[\xca\xe6\xe8\x84+\xa1\xc66", {"Content-Encoding": "br"}),
        ],
        ids=["plain", "gzip", "deflate", "zstd", "br"],
    )
    @pytest.mark.asyncio
    async def test_compressing_support(
        self, respx_mock: MockRouter, bytes_content: bytes, expected_headers: dict[str, str]
    ) -> None:
        expected_content = "test-response"

        mocked_response = httpx.Response(200, headers=expected_headers, content=bytes_content)

        respx_mock.get("/test").mock(return_value=mocked_response)

        response = await load_content(
            url=str_to_absolute_url("http://example.com/test"), proxy=Proxy(name="test", url=None), user_agent="test"
        )

        assert response.text == expected_content

    @pytest.mark.asyncio
    async def test_accept_encoding_header(self, respx_mock: MockRouter) -> None:
        respx_mock.get("/test").mock()

        await load_content(
            url=str_to_absolute_url("http://example.com/test"), proxy=Proxy(name="test", url=None), user_agent="test"
        )

        assert (
            respx_mock.calls[0].request.headers["Accept-Encoding"] == "br;q=1.0, zstd;q=0.9, gzip;q=0.8, deflate;q=0.7"
        )

    @pytest.mark.asyncio
    async def test_expected_error(self, respx_mock: MockRouter) -> None:
        respx_mock.get("/test").mock(side_effect=httpx.ConnectTimeout("some message"))

        with capture_logs() as logs:
            with pytest.raises(errors.LoadError) as expected_error:
                await load_content(
                    url=str_to_absolute_url("http://example.com/test"),
                    proxy=Proxy(name="test", url=None),
                    user_agent="test",
                )

        assert expected_error.value.feed_error_code == FeedError.network_connection_timeout

        assert_logs(logs, error_while_loading_feed=0)
        assert_logs_have_no_errors(logs)

    @pytest.mark.asyncio
    async def test_unexpected_error(self, respx_mock: MockRouter) -> None:
        respx_mock.get("/test").mock(side_effect=Exception("some message"))

        with capture_logs() as logs:
            with pytest.raises(errors.LoadError) as expected_error:
                await load_content(
                    url=str_to_absolute_url("http://example.com/test"),
                    proxy=Proxy(name="test", url=None),
                    user_agent="test",
                )

        assert expected_error.value.feed_error_code == FeedError.network_unknown

        assert_logs(logs, error_while_loading_feed=1)
        assert_logs_levels(logs, error_while_loading_feed="error")



import uuid
from typing import Any

from ffun.domain.entities import FeedId
from ffun.domain.urls import str_to_absolute_url, str_to_feed_url, url_to_source_uid
from ffun.feeds.entities import Feed, FeedState
from ffun.feeds.operations import get_feeds, get_source_ids, save_feed


def fake_url() -> str:
    return f"https://{uuid.uuid4().hex}.com"


def fake_title() -> str:
    return f"Feed Title: {uuid.uuid4().hex}"


def fake_description() -> str:
    return f"Feed Description: {uuid.uuid4().hex}"


async def fake_feed(**kwargs: Any) -> Feed:
    url = fake_url() if "url" not in kwargs else kwargs["url"]

    source_uid = url_to_source_uid(str_to_absolute_url(url))

    source_ids = await get_source_ids([source_uid])

    return Feed(
        id=FeedId(uuid.uuid4() if "id" not in kwargs else kwargs["id"]),
        source_id=source_ids[source_uid],
        url=str_to_feed_url(url),
        state=FeedState.not_loaded if "state" not in kwargs else kwargs["state"],
        last_error=kwargs.get("last_error"),
        load_attempted_at=kwargs.get("load_attempted_at"),
        loaded_at=kwargs.get("loaded_at"),
        title=fake_title() if "title" not in kwargs else kwargs["title"],
        description=fake_description() if "description" not in kwargs else kwargs["description"],
    )


async def n_feeds(n: int) -> list[Feed]:
    feed_objects = [await fake_feed() for _ in range(n)]

    saved_ids = [await save_feed(feed) for feed in feed_objects]

    feeds = await get_feeds(saved_ids)

    return feeds



import asyncio
import pathlib
from collections import Counter

import typer

from ffun.application.application import with_app
from ffun.core import logging
from ffun.processors_quality import domain as pq_domain
from ffun.processors_quality.domain import diff_processor_results, display_diffs
from ffun.processors_quality.entities import ProcessorResult
from ffun.processors_quality.knowlege_base import KnowlegeBase

logger = logging.get_module_logger()

cli_app = typer.Typer()


_root = pathlib.Path("./tags_quality_base/")


async def single_run(processor_name: str, entry_id: int, kb: KnowlegeBase, actual: bool = False) -> ProcessorResult:
    logger.info("processing_entry", entry_id=entry_id)
    result = await pq_domain.run_processor(kb, processor_name, entry_id)
    kb.save_processor_result(processor_name, entry_id, result, actual=actual)
    return result


async def run_one(
    processor_name: str, entry_id: int, knowlege_root: pathlib.Path, actual: bool, show_tag_diffs: bool
) -> None:
    async with with_app():
        kb = KnowlegeBase(knowlege_root)

        await single_run(processor_name, entry_id, kb, actual=actual)

        diffs = diff_processor_results(kb, processor_name, [entry_id])

        display_diffs(diffs, show_tag_diffs=show_tag_diffs)


@cli_app.command()
def test_one(
    processor: str, entry: int, knowlege_root: pathlib.Path = _root, actual: bool = False, show_tag_diffs: bool = False
) -> None:
    asyncio.run(run_one(processor, entry, knowlege_root, actual=actual, show_tag_diffs=show_tag_diffs))


async def run_all(processor_name: str, knowlege_root: pathlib.Path, actual: bool, show_tag_diffs: bool) -> None:
    async with with_app():
        kb = KnowlegeBase(knowlege_root)

        ids = kb.entry_ids()

        tasks = []

        for entry_id in ids:
            tasks.append(single_run(processor_name, entry_id, kb, actual=actual))

        await asyncio.gather(*tasks)

        diffs = diff_processor_results(kb, processor_name, ids)

        display_diffs(diffs, show_tag_diffs=show_tag_diffs)


@cli_app.command()
def test_all(
    processor: str, knowlege_root: pathlib.Path = _root, actual: bool = False, show_tag_diffs: bool = False
) -> None:
    asyncio.run(run_all(processor, knowlege_root, actual=actual, show_tag_diffs=show_tag_diffs))


async def run_validate_expected_tags(knowlege_root: pathlib.Path = _root) -> None:
    async with with_app():
        kb = KnowlegeBase(knowlege_root)
        kb.validate_expected_tags()


@cli_app.command()
def validate_expected_tags(knowlege_root: pathlib.Path = _root) -> None:
    asyncio.run(run_validate_expected_tags(knowlege_root=knowlege_root))


async def run_copy_last_to_actual(processor: str, entry: int, knowlege_root: pathlib.Path = _root) -> None:
    async with with_app():
        kb = KnowlegeBase(knowlege_root)
        kb.copy_last_to_actual(processor, entry)


@cli_app.command()
def copy_last_to_actual(processor: str, entry: int, knowlege_root: pathlib.Path = _root) -> None:
    asyncio.run(run_copy_last_to_actual(processor, entry, knowlege_root=knowlege_root))


async def run_copy_last_to_actual_all(processor: str, knowlege_root: pathlib.Path = _root) -> None:
    async with with_app():
        kb = KnowlegeBase(knowlege_root)
        ids = kb.entry_ids()

        for i, entry_id in enumerate(ids):
            logger.info("copying_last_to_actual", entry_id=entry_id, i=i, total=len(ids))
            kb.copy_last_to_actual(processor, entry_id)


@cli_app.command()
def copy_last_to_actual_all(processor: str, knowlege_root: pathlib.Path = _root) -> None:
    asyncio.run(run_copy_last_to_actual_all(processor, knowlege_root=knowlege_root))


async def run_diff_entry(
    processor_name: str, entry_id: int, show_tag_diffs: bool, knowlege_root: pathlib.Path = _root
) -> None:
    async with with_app():
        kb = KnowlegeBase(knowlege_root)

        diffs = diff_processor_results(kb, processor_name, [entry_id])

        display_diffs(diffs, show_tag_diffs=show_tag_diffs)


@cli_app.command()
def diff_entry(processor: str, entry: int, knowlege_root: pathlib.Path = _root, show_tag_diffs: bool = False) -> None:
    asyncio.run(run_diff_entry(processor, entry, knowlege_root=knowlege_root, show_tag_diffs=show_tag_diffs))


async def run_deff_all(processor_name: str, knowlege_root: pathlib.Path = _root, show_tag_diffs: bool = False) -> None:
    async with with_app():
        kb = KnowlegeBase(knowlege_root)

        diffs = diff_processor_results(kb, processor_name, kb.entry_ids())

        display_diffs(diffs, show_tag_diffs)


@cli_app.command()
def diff_all(processor: str, knowlege_root: pathlib.Path = _root, show_tag_diffs: bool = False) -> None:
    asyncio.run(run_deff_all(processor, knowlege_root=knowlege_root, show_tag_diffs=show_tag_diffs))


async def run_prepare_news_item(
    processor: str, entry_id: int, knowlege_root: pathlib.Path, requests_number: int, min_tags_count: int
) -> None:
    results = []

    async with with_app():
        kb = KnowlegeBase(knowlege_root)

        for i in range(requests_number):
            logger.info("requesting_tags", step=i, steps_number=requests_number)
            result = await single_run(processor, entry_id, kb, actual=True)
            results.append(result)

        logger.info("requests_completed")

        tags: Counter[str] = Counter()

        for result in results:
            tags.update(result.tags)

        tags_must_have = {tag for tag, count in tags.items() if count == requests_number}
        tags_should_have = {tag for tag, count in tags.items() if min_tags_count <= count < requests_number}

        kb.save_expected_data(processor, entry_id, tags_must_have=tags_must_have, tags_should_have=tags_should_have)


@cli_app.command()
def prepere_news_item(
    processor: str, entry: int, knowlege_root: pathlib.Path = _root, requests_number: int = 10, min_tags_count: int = 5
) -> None:
    asyncio.run(
        run_prepare_news_item(
            processor,
            entry_id=entry,
            knowlege_root=knowlege_root,
            requests_number=requests_number,
            min_tags_count=min_tags_count,
        )
    )



import pydantic

from ffun.core.entities import BaseEntity
from ffun.domain.entities import CollectionId, CollectionSlug, FeedUrl
from ffun.feeds.entities import FeedId


class FeedInfo(BaseEntity):
    model_config = pydantic.ConfigDict(frozen=False)

    url: FeedUrl

    # Some feeds define title and description in their data
    # But, currently, it looks more convenient to define them here
    # So, we can describe feeds in a consistent way resistent to problems on the feed side
    title: str
    description: str

    feed_id: FeedId | None = None


class Collection(BaseEntity):
    id: CollectionId
    slug: CollectionSlug
    gui_order: int
    name: str
    description: str
    feeds: list[FeedInfo]
    show_on_main: bool



from typing import Any

from psycopg import Connection
from yoyo import step

__depends__: set[str] = set()


sql_create_feeds_table = """
CREATE TABLE f_feeds (
    id UUID PRIMARY KEY,
    url TEXT NOT NULL UNIQUE,
    state INTEGER NOT NULL,
    last_error INTEGER NULL DEFAULT NULL,
    loaded_at TIMESTAMP WITH TIME ZONE NULL DEFAULT NULL,
    load_attempted_at TIMESTAMP WITH TIME ZONE NULL DEFAULT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW(),
    updated_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW())
"""

sql_create_load_attempted_at_index = """
CREATE INDEX idx_f_feeds_load_attempted_at_at ON f_feeds (load_attempted_at);
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_feeds_table)
    cursor.execute(sql_create_load_attempted_at_index)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE f_feeds")


steps = [step(apply_step, rollback_step)]



import uuid
from typing import Any

from ffun.core import utils
from ffun.domain.domain import new_entry_id
from ffun.domain.entities import EntryId, SourceId
from ffun.domain.urls import str_to_absolute_url
from ffun.feeds.entities import Feed
from ffun.library import domain
from ffun.library.entities import Entry


def fake_url() -> str:
    return f"https://{uuid.uuid4().hex}.com"


def fake_title() -> str:
    return f"Entry Title: {uuid.uuid4().hex}"


def fake_body() -> str:
    return f"Entry Body: {uuid.uuid4().hex}"


def fake_entry(source_id: SourceId, **kwargs: Any) -> Entry:
    return Entry(
        id=new_entry_id() if "id" not in kwargs else kwargs["id"],
        source_id=source_id,
        title=fake_title() if "title" not in kwargs else kwargs["title"],
        body=fake_body() if "body" not in kwargs else kwargs["body"],
        external_id=uuid.uuid4().hex if "external_id" not in kwargs else kwargs["external_id"],
        external_url=str_to_absolute_url(fake_url() if "external_url" not in kwargs else kwargs["external_url"]),
        external_tags={uuid.uuid4().hex, uuid.uuid4().hex},
        published_at=utils.now() if "published_at" not in kwargs else kwargs["published_at"],
        cataloged_at=utils.now() if "cataloged_at" not in kwargs else kwargs["cataloged_at"],
    )


async def n_entries(feed: Feed, n: int) -> dict[EntryId, Entry]:
    new_entries = [fake_entry(feed.source_id) for _ in range(n)]
    await domain.catalog_entries(feed.id, new_entries)
    return await domain.get_entries_by_ids([entry.id for entry in new_entries])  # type: ignore


async def n_entries_list(feed: Feed, n: int) -> list[Entry]:
    entries = await n_entries(feed, n)

    result = list(entries.values())

    # from newest to oldest
    result.sort(key=lambda entry: entry.cataloged_at, reverse=True)

    return result



import typer

from ffun.cli.commands import cleaner  # noqa: F401
from ffun.cli.commands import estimates  # noqa: F401
from ffun.cli.commands import experiments  # noqa: F401
from ffun.cli.commands import fixtures  # noqa: F401
from ffun.cli.commands import metrics  # noqa: F401
from ffun.cli.commands import processors_quality  # noqa: F401
from ffun.cli.commands import profile  # noqa: F401
from ffun.cli.commands import user_settings  # noqa: F401
from ffun.core import logging

app = typer.Typer()

logger = logging.get_module_logger()


app.add_typer(processors_quality.cli_app, name="processors-quality")
app.add_typer(user_settings.cli_app, name="user-settings")
app.add_typer(cleaner.cli_app, name="cleaner")
app.add_typer(metrics.cli_app, name="metrics")
app.add_typer(profile.cli_app, name="profile")
app.add_typer(experiments.cli_app, name="experiments")
app.add_typer(fixtures.cli_app, name="fixtures")
app.add_typer(estimates.cli_app, name="estimates")


if __name__ == "__main__":
    app()



import pytest
from pytest_mock import MockerFixture

from ffun.core.tests.helpers import assert_logs_has_business_event, assert_logs_has_no_business_event, capture_logs
from ffun.data_protection.domain import remove_user
from ffun.domain.domain import new_user_id
from ffun.domain.entities import UserId
from ffun.users.domain import get_user_external_ids


# TODO: add tests for multiple services when we'll support them
class TestRemoveUser:

    @pytest.mark.asyncio
    async def test_no_user(self) -> None:
        user_id = new_user_id()

        with capture_logs() as logs:
            await remove_user(user_id)

        assert_logs_has_no_business_event(logs, "user_removed")

    @pytest.mark.asyncio
    async def test_has_user_single_service(self, internal_user_id: UserId, mocker: MockerFixture) -> None:

        remove_user_from_external_service = mocker.patch("ffun.auth.domain.remove_user_from_external_service")

        assert await get_user_external_ids(internal_user_id)

        with capture_logs() as logs:
            await remove_user(internal_user_id)

        assert remove_user_from_external_service.call_count == 1

        assert not await get_user_external_ids(internal_user_id)

        assert_logs_has_business_event(logs, "user_removed", user_id=internal_user_id)



import asyncio
from typing import Any

from ffun.core import logging
from ffun.core.background_tasks import InfiniteTask
from ffun.domain.entities import EntryId
from ffun.feeds_collections.collections import collections
from ffun.librarian import domain, operations
from ffun.librarian.entities import ProcessorType
from ffun.librarian.processors.base import Processor
from ffun.librarian.processors.domain import Processor as DomainProcessor
from ffun.librarian.processors.llm_general import Processor as LLMGeneralProcessor
from ffun.librarian.processors.native_tags import Processor as NativeTagsProcessor
from ffun.librarian.processors.upper_case_title import Processor as UpperCaseTitleProcessor
from ffun.librarian.settings import settings
from ffun.library import domain as l_domain
from ffun.library.entities import Entry

logger = logging.get_module_logger()


# TODO: merge with EntriesProcessor or with base.Processor?
class ProcessorInfo:
    __slots__ = ("id", "processor", "concurrency", "type", "allowed_for_collections", "allowed_for_users")

    def __init__(  # noqa: disable=CFQ002
        self,
        id: int,
        type: ProcessorType,
        processor: Processor,
        concurrency: int,
        allowed_for_collections: bool,
        allowed_for_users: bool,
    ):
        self.type = type
        self.id = id
        self.processor = processor
        self.concurrency = concurrency
        self.allowed_for_collections = allowed_for_collections
        self.allowed_for_users = allowed_for_users


processors: list[ProcessorInfo] = []


for processor_config in settings.tag_processors:
    if not processor_config.enabled:
        logger.info(
            "tag_processor_is_disabled", processor_id=processor_config.id, processor_name=processor_config.name
        )
        continue

    logger.info("add_tag_processor", processor_id=processor_config.id, processor_name=processor_config.name)

    info_arguments = {}

    tags_processor: Processor

    if processor_config.type == ProcessorType.domain:
        tags_processor = DomainProcessor(name=processor_config.name)
    elif processor_config.type == ProcessorType.native_tags:
        tags_processor = NativeTagsProcessor(name=processor_config.name)
    elif processor_config.type == ProcessorType.upper_case_title:
        tags_processor = UpperCaseTitleProcessor(name=processor_config.name)
    elif processor_config.type == ProcessorType.llm_general:
        tags_processor = LLMGeneralProcessor(
            name=processor_config.name,
            entry_template=processor_config.entry_template,
            text_cleaner=processor_config.text_cleaner,
            tag_extractor=processor_config.tags_extractor,
            llm_provider=processor_config.llm_provider,
            llm_config=processor_config.llm_config,
            collections_api_key=processor_config.collections_api_key,
            general_api_key=processor_config.general_api_key,
        )
    else:
        raise NotImplementedError(f"Unknown processor type: {processor_config.type}")

    info_arguments["processor"] = tags_processor

    processor = ProcessorInfo(
        id=processor_config.id,
        type=processor_config.type,
        processor=tags_processor,
        concurrency=processor_config.workers,
        allowed_for_collections=processor_config.allowed_for_collections,
        allowed_for_users=processor_config.allowed_for_users,
    )

    processors.append(processor)


class EntriesProcessor(InfiniteTask):
    __slots__ = ("_processor_info",)

    def __init__(self, processor_info: ProcessorInfo, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._processor_info = processor_info

    @property
    def id(self) -> int:
        return self._processor_info.id

    @property
    def concurrency(self) -> int:
        return self._processor_info.concurrency

    async def separate_entries(  # noqa: disable=CCR001
        self, entries_ids: list[EntryId]
    ) -> tuple[list[Entry], list[EntryId]]:
        processor_id = self._processor_info.id

        # TODO: we could add caching here, because multiple processors can request the same entries
        #       and currently we run multiple processors in the same process
        #       but remember that for API worker such cachin may be useless or even harmful
        entries = await l_domain.get_entries_by_ids(entries_ids)

        entries_to_process: list[Entry] = []
        entries_to_remove: list[EntryId] = []

        feed_links = await l_domain.get_feed_links_for_entries(entries_ids)
        feed_ids = {entry_id: [link.feed_id for link in links] for entry_id, links in feed_links.items()}

        for entry_id, entry in entries.items():
            if entry is None:
                logger.warning("unexisted_entry_in_queue", processor_id=processor_id, entry_id=entry_id)
                entries_to_remove.append(entry_id)
                continue

            if entry_id not in feed_ids:
                # this may happen when:
                # - we clean up DB from old data => drop linking, mark entries as orphaned
                # - but we still have these orphaned entries in the queues
                # - because the processor was stopped temporary or failed by other means
                logger.warning("entry_without_feeds_in_queue", processor_id=processor_id, entry_id=entry_id)
                entries_to_remove.append(entry_id)
                continue

            # TODO: maybe we should create some `in_collection` marker directly for entry
            #       this may simplify a lot of code by moveing checks from multiple places to one
            in_collection = any(collections.has_feed(feed_id) for feed_id in feed_ids[entry_id])

            if in_collection and not self._processor_info.allowed_for_collections:
                logger.info("proccessor_not_allowed_for_collections", processor_id=processor_id, entry_id=entry_id)
                entries_to_remove.append(entry_id)
                continue

            if not in_collection and not self._processor_info.allowed_for_users:
                logger.info("proccessor_not_allowed_for_users", processor_id=processor_id, entry_id=entry_id)
                entries_to_remove.append(entry_id)
                continue

            logger.info("proccessor_is_allowed_for_entry", processor_id=processor_id, entry_id=entry_id)
            entries_to_process.append(entry)

        return entries_to_process, entries_to_remove

    async def single_run(self) -> None:
        processor_id = self._processor_info.id
        concurrency = self._processor_info.concurrency

        # most likely, this call should be in a separate worker with a more complex logic
        # but for now it is ok to place it here
        await domain.plan_processor_queue(
            processor_id=processor_id,
            fill_when_below=concurrency,
            # TODO: move to settings
            chunk=concurrency * 10,
        )

        entries_ids = await operations.get_entries_to_process(processor_id=processor_id, limit=concurrency)

        if not entries_ids:
            logger.info("no_entries_to_process", processor_id=processor_id)
            return

        entries_to_process, entries_to_remove = await self.separate_entries(entries_ids=entries_ids)

        tasks = []

        for entry in entries_to_process:
            tasks.append(
                domain.process_entry(processor_id=processor_id, processor=self._processor_info.processor, entry=entry)
            )

        if entries_to_remove:
            tasks.append(
                domain.remove_entries_from_processor_queue(processor_id=processor_id, entry_ids=entries_to_remove)
            )

        await asyncio.gather(*tasks, return_exceptions=True)


def create_background_processors() -> list[EntriesProcessor]:
    return [
        EntriesProcessor(
            processor_info=processor_info,
            name=f"entries_processor_{processor_info.processor.name}",
            delay_between_runs=1,
        )
        for processor_info in processors
    ]



import datetime

import pydantic

from ffun.core.entities import BaseEntity


class ExpectedTags(BaseEntity):
    must_have: set[str]
    should_have: set[str]

    @pydantic.model_validator(mode="after")
    def tags_must_not_intersect(self) -> "ExpectedTags":
        if self.must_have & self.should_have:
            raise ValueError(f"Tags must not intersect, common tags: {self.must_have & self.should_have}")

        return self


class ProcessorResult(BaseEntity):
    tags: list[str]
    created_at: datetime.datetime


class ProcessorResultDiff(BaseEntity):
    entry_id: int

    actual_total: int
    actual_must_have_found: int
    actual_must_have_missing: list[str]
    actual_should_have_found: int
    actual_has_and_last_not: list[str]

    last_total: int
    last_must_have_found: int
    last_must_have_missing: list[str]
    last_should_have_found: int
    last_has_and_actual_not: list[str]

    must_have_total: int
    should_have_total: int

    @pydantic.field_validator("actual_must_have_missing")
    @classmethod
    def sort_actual_must_have_missing_tags(cls, v: list[str]) -> list[str]:
        return list(sorted(v))

    @pydantic.field_validator("last_must_have_missing")
    @classmethod
    def sort_last_must_have_missing_tags(cls, v: list[str]) -> list[str]:
        return list(sorted(v))

    @pydantic.field_validator("actual_has_and_last_not")
    @classmethod
    def sort_actual_has_and_last_not_tags(cls, v: list[str]) -> list[str]:
        return list(sorted(v))

    @pydantic.field_validator("last_has_and_actual_not")
    @classmethod
    def sort_last_has_and_actual_not_tags(cls, v: list[str]) -> list[str]:
        return list(sorted(v))


class Attribution(BaseEntity):
    title: str = pydantic.Field(..., min_length=1)
    authors: list[str] = pydantic.Field(..., min_items=1)  # type: ignore
    link: str = pydantic.Field(..., min_length=1)
    license: str = pydantic.Field(..., min_length=1)



import enum
from typing import Any

import pydantic

from ffun.core.entities import BaseEntity


class APIStatuses(str, enum.Enum):
    success = "success"
    error = "error"


class MessageType(str, enum.Enum):
    info = "info"
    warning = "warning"
    error = "error"


class Message(BaseEntity):
    type: MessageType
    code: str
    message: str


class APIRequest(BaseEntity):
    model_config = pydantic.ConfigDict(title="Request Body")


class APISuccess(BaseEntity):
    status: APIStatuses = APIStatuses.success
    messages: list[Message] = pydantic.Field(default_factory=list)

    model_config = pydantic.ConfigDict(title="Response Body")


# TODO: refactor to allow list of messages?
class APIError(BaseEntity):
    status: APIStatuses = APIStatuses.error
    code: str
    message: str = "Unknown error"
    data: dict[str, Any] | None = None

    model_config = pydantic.ConfigDict(title="API Error")



import pytest

from ffun.tags.converters import _decode_special_characters, _encode_special_characters, normalize


class TestEncodeSpecialCharacters:
    @pytest.mark.parametrize(
        "tag, expected",
        [
            ("abc", "abc"),
            ("abc def", "abc def"),
            ("c++", "c-plus--plus-"),
            ("c#", "c-sharp-"),
            ("www.example.com", "www-dot-example-dot-com"),
        ],
    )
    def test(self, tag: str, expected: str) -> None:
        assert _encode_special_characters(tag) == expected


class TestNormalize:
    @pytest.mark.parametrize(
        "tag, expected",
        [
            ("abc", "abc"),
            ("abc def", "abc-def"),
            ("c++", "c-plus-plus"),
            ("c#", "c-sharp"),
            ("www.example.com", "www-dot-example-dot-com"),
        ],
    )
    def test(self, tag: str, expected: str) -> None:
        assert normalize(tag) == expected


class TestDecodeSpecialCharacters:
    @pytest.mark.parametrize(
        "tag, expected",
        [
            ("abc", "abc"),
            ("abc-def", "abc-def"),
            ("c-plus-plus", "c++"),
            ("c-sharp", "c#"),
            ("www-dot-example-dot-com", "www.example.com"),
        ],
    )
    def test(self, tag: str, expected: str) -> None:
        assert _decode_special_characters(tag) == expected



import asyncio
import contextlib
import functools
from typing import Any, AsyncGenerator, Callable, Protocol

import psycopg
import psycopg_pool
from psycopg.pq import ExecStatus
from psycopg.rows import dict_row

from ffun.core import logging

logger = logging.get_module_logger()

MAX_INTEGER = 2147483647

POOL: psycopg_pool.AsyncConnectionPool | None = None


DB_RESULT = list[dict[str, Any]]
SQL_ARGUMENTS = dict[str, Any] | tuple[list[Any]]


class ExecuteType(Protocol):
    async def __call__(self, command: str, arguments: SQL_ARGUMENTS | None = None) -> DB_RESULT:
        pass


class PGAsyncCursor(psycopg.AsyncCursor):
    async def execute_and_extract(self, command: str, arguments: SQL_ARGUMENTS | None = None) -> DB_RESULT:
        await self.execute(command, arguments)

        if self.pgresult is None:
            raise NotImplementedError("We expect cursor.pgresult has already settled here")

        # see for details
        # https://github.com/psycopg/psycopg/blob/ea76ab81ba1d797eee2baf2a1464be51e608b8bd/psycopg/psycopg/pq/_enums.py
        if self.pgresult.status in (ExecStatus.TUPLES_OK, ExecStatus.SINGLE_TUPLE):
            return await self.fetchall()  # type: ignore

        return []


class PGPAsyncConnection(psycopg.AsyncConnection):
    def __init__(self, *argv, row_factory=dict_row, cursor_factory=PGAsyncCursor, **kwargs):  # type: ignore
        super().__init__(*argv, row_factory=row_factory, **kwargs)  # type: ignore
        self.cursor_factory = cursor_factory

        # set prepare_threshold to None because we can use connection pool (pgbouncer/RDSPRoxy)
        # details: https://www.psycopg.org/psycopg3/docs/advanced/prepare.html
        self.prepare_threshold = None


async def pool_refresher(delay: int) -> None:
    logger.info("star_pool_refresher")

    try:
        while True:
            await asyncio.sleep(delay)

            if POOL is None:
                continue

            logger.info("refresh_pool")

            await POOL.check()
    except asyncio.CancelledError:
        logger.info("pool_refresher_is_stopped")
        return


async def prepare_pool(  # noqa: CFQ002
    name: str, dsn: str, min_size: int, max_size: int | None, timeout: float, num_workers: int, max_lifetime: int
) -> None:
    global POOL

    if POOL is not None:
        raise RuntimeError("Secondary db pool initialization is not allowed")

    POOL = psycopg_pool.AsyncConnectionPool(
        dsn,
        min_size=min_size,
        max_size=max_size,
        max_lifetime=max_lifetime,
        timeout=timeout,
        connection_class=PGPAsyncConnection,
        num_workers=num_workers,
        name=name,
        open=False,
        kwargs={"autocommit": True},
    )

    await POOL.open(wait=False)


async def destroy_pool() -> None:
    global POOL

    if POOL is None:
        return

    try:
        await POOL.close()
    except asyncio.CancelledError:
        # TODO: is this a correct solution?
        pass

    POOL = None


@contextlib.asynccontextmanager
async def transaction(autocommit: bool = False) -> AsyncGenerator[ExecuteType, None]:
    if POOL is None:
        raise RuntimeError("POOL MUST be initialized before any operations with database")

    async with POOL.connection() as connection:
        if connection.autocommit != autocommit:
            await connection.set_autocommit(autocommit)

        async with connection.cursor() as cursor:
            assert isinstance(cursor, PGAsyncCursor)
            yield cursor.execute_and_extract


async def execute(command: str, arguments: SQL_ARGUMENTS | None = None) -> DB_RESULT:
    async with transaction(autocommit=True) as execute:
        return await execute(command, arguments)


def run_in_transaction(func: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(func)
    async def wrapper(*argv: Any, **kwargs: Any) -> Any:
        async with transaction() as execute:
            return await func(execute, *argv, **kwargs)

    return wrapper



import pytest

from ffun.domain.entities import FeedId, UserId
from ffun.feeds_links.domain import get_linked_users_flat
from ffun.feeds_links.operations import add_link


class TestGetLinkedUsersFlat:

    @pytest.mark.asyncio
    async def test_empty(self, saved_feed_id: FeedId) -> None:
        users = await get_linked_users_flat([saved_feed_id])

        assert users == set()

    @pytest.mark.asyncio
    async def test_returns_only_required(
        self, five_internal_user_ids: list[UserId], saved_feed_id: FeedId, another_saved_feed_id: FeedId
    ) -> None:
        user_1_id, user_2_id, user_3_id = five_internal_user_ids[:3]

        await add_link(user_1_id, saved_feed_id)

        await add_link(user_2_id, saved_feed_id)
        await add_link(user_2_id, another_saved_feed_id)

        await add_link(user_3_id, another_saved_feed_id)

        users = await get_linked_users_flat([saved_feed_id])

        assert users == {user_1_id, user_2_id}

        users = await get_linked_users_flat([another_saved_feed_id])

        assert users == {user_2_id, user_3_id}

        users = await get_linked_users_flat([saved_feed_id, another_saved_feed_id])

        assert users == {user_1_id, user_2_id, user_3_id}



import datetime

import pytest
import pytest_asyncio

from ffun.core import utils
from ffun.feeds.entities import Feed
from ffun.library import operations
from ffun.library.domain import get_entries_by_filter_with_fallback, get_entry, get_feeds_for_entry, normalize_entry
from ffun.library.entities import Entry, EntryChange
from ffun.library.tests import helpers, make


class TestNormalizeEntry:
    @pytest.mark.asyncio
    async def test_no_changes(self, cataloged_entry: Entry) -> None:
        changes = await normalize_entry(cataloged_entry)
        assert changes == []

        loaded_entry = await get_entry(cataloged_entry.id)
        assert loaded_entry == cataloged_entry

    @pytest.mark.parametrize("apply", [True, False])
    @pytest.mark.asyncio
    async def test_normalize_external_url(
        self, new_entry: Entry, loaded_feed: Feed, another_loaded_feed: Feed, apply: bool
    ) -> None:
        wrong_url = "/relative/url"
        expected_url = f"{loaded_feed.url}{wrong_url}"

        new_entry = new_entry.replace(external_url=wrong_url)

        assert loaded_feed.source_id == new_entry.source_id

        await operations.catalog_entries(loaded_feed.id, [new_entry])

        # this feed should be ignored, because current logic uses only the oldest link
        await operations.catalog_entries(another_loaded_feed.id, [new_entry])

        entry = await get_entry(new_entry.id)

        changes = await normalize_entry(entry, apply=apply)

        assert changes == [
            EntryChange(
                id=new_entry.id, field="external_url", old_value=new_entry.external_url, new_value=expected_url
            )
        ]

        loaded_entry = await get_entry(new_entry.id)

        if apply:
            assert loaded_entry.external_url == expected_url
        else:
            assert loaded_entry.external_url == wrong_url


class TestGetFeedsForEntry:

    @pytest.mark.asyncio
    async def test_single_feed(self, new_entry: Entry, loaded_feed: Feed) -> None:
        await operations.catalog_entries(loaded_feed.id, [new_entry])

        feeds = await get_feeds_for_entry(new_entry.id)

        assert feeds == {loaded_feed.id}

    @pytest.mark.asyncio
    async def test_multiple_feed(self, new_entry: Entry, loaded_feed: Feed, another_loaded_feed: Feed) -> None:
        await operations.catalog_entries(loaded_feed.id, [new_entry])
        await operations.catalog_entries(another_loaded_feed.id, [new_entry])

        feeds = await get_feeds_for_entry(new_entry.id)

        assert feeds == {loaded_feed.id, another_loaded_feed.id}

    @pytest.mark.asyncio
    async def test_no_feeds(self, new_entry: Entry, loaded_feed: Feed) -> None:
        await operations.catalog_entries(loaded_feed.id, [new_entry])

        await operations.unlink_feed_tail(loaded_feed.id, 0)

        feeds = await get_feeds_for_entry(new_entry.id)

        assert feeds == set()


# get_entries_by_filter_with_fallback uses get_entries_by_filter
# so, we may test only fallback logic
class TestGetEntriesByFilterWithFallback:

    @pytest.fixture
    def time_delta(self) -> datetime.timedelta:
        return datetime.timedelta(days=1)

    @pytest.fixture
    def time_border(self, time_delta: datetime.timedelta) -> datetime.datetime:
        return utils.now() - time_delta

    @pytest_asyncio.fixture
    async def prepared_entries(self, loaded_feed: Feed, time_border: datetime.datetime) -> list[Entry]:
        entries = await make.n_entries_list(loaded_feed, n=3)

        await helpers.update_cataloged_time(
            entries_ids=[entry.id for entry in entries], new_time=time_border - datetime.timedelta(seconds=10)
        )

        all_entries = await operations.get_entries_by_ids(ids=[entry.id for entry in entries])

        all_entries_list = [entry for entry in all_entries.values() if entry is not None]
        all_entries_list.sort(key=lambda entry: entry.cataloged_at)

        return all_entries_list

    @pytest.mark.asyncio
    async def test_no_entries_at_all(self, time_delta: datetime.timedelta) -> None:
        entries = await get_entries_by_filter_with_fallback(
            feeds_ids=[], period=time_delta, limit=10, fallback_limit=10
        )

        assert entries == []

    @pytest.mark.asyncio
    async def test_has_new_entries(
        self, loaded_feed: Feed, time_border: datetime.datetime, time_delta: datetime.timedelta
    ) -> None:
        entries = await make.n_entries_list(loaded_feed, n=3)

        await helpers.update_cataloged_time(
            entries_ids=[entries[-1].id], new_time=time_border - datetime.timedelta(seconds=10)
        )

        loaded_entries = await get_entries_by_filter_with_fallback(
            feeds_ids=[loaded_feed.id], period=time_delta, limit=10, fallback_limit=10
        )

        assert {entry.id for entry in loaded_entries} == {entry.id for entry in entries if entry.id != entries[-1].id}

    @pytest.mark.asyncio
    async def test_has_new_entries__limit(
        self, loaded_feed: Feed, time_border: datetime.datetime, time_delta: datetime.timedelta
    ) -> None:
        entries = await make.n_entries_list(loaded_feed, n=3)

        await helpers.update_cataloged_time(
            entries_ids=[entries[-1].id], new_time=time_border - datetime.timedelta(seconds=10)
        )

        loaded_entries = await get_entries_by_filter_with_fallback(
            feeds_ids=[loaded_feed.id], period=time_delta, limit=1, fallback_limit=10
        )

        assert len(loaded_entries) == 1
        assert loaded_entries[0].id in {entry.id for entry in entries if entry.id != entries[-1].id}

    @pytest.mark.asyncio
    async def test_no_new_entries(
        self, loaded_feed: Feed, time_border: datetime.datetime, time_delta: datetime.timedelta
    ) -> None:
        entries = await make.n_entries_list(loaded_feed, n=3)

        await helpers.update_cataloged_time(
            entries_ids=[entry.id for entry in entries], new_time=time_border - datetime.timedelta(seconds=10)
        )

        loaded_entries = await get_entries_by_filter_with_fallback(
            feeds_ids=[loaded_feed.id], period=time_delta, limit=1, fallback_limit=10
        )

        assert {entry.id for entry in loaded_entries} == {entry.id for entry in entries}

    @pytest.mark.asyncio
    async def test_no_new_entries__limit(
        self, loaded_feed: Feed, time_border: datetime.datetime, time_delta: datetime.timedelta
    ) -> None:
        entries = await make.n_entries_list(loaded_feed, n=3)

        await helpers.update_cataloged_time(
            entries_ids=[entry.id for entry in entries], new_time=time_border - datetime.timedelta(seconds=10)
        )

        loaded_entries = await get_entries_by_filter_with_fallback(
            feeds_ids=[loaded_feed.id], period=time_delta, limit=1, fallback_limit=1
        )

        assert len(loaded_entries) == 1
        assert loaded_entries[0].id in {entry.id for entry in entries}



import asyncio

import typer

from ffun.application.application import with_app
from ffun.core import logging
from ffun.core.postgresql import execute
from ffun.domain.entities import UnknownUrl
from ffun.domain.urls import normalize_classic_unknown_url, url_to_source_uid, url_to_uid
from ffun.library.operations import all_entries_iterator, count_total_entries

logger = logging.get_module_logger()

cli_app = typer.Typer()


async def run_check_entries() -> None:
    async with with_app():

        logger.info("start_experiment")

        counter = 0

        chunk = 10000

        total = await count_total_entries()

        wrong_urls = 0

        async for entry in all_entries_iterator(chunk=chunk):
            counter += 1

            if counter % chunk == 0:
                logger.info(
                    "processed_entries",
                    counter=counter,
                    percentage=round((counter / total) * 100, 2),
                    wrong_urls=wrong_urls,
                )

            if "#" in entry.external_url or "magnet" in entry.external_url:
                continue

            if normalize_classic_unknown_url(UnknownUrl(entry.external_url)) != entry.external_url:
                wrong_urls += 1

        logger.info("experiment_finished", wrong_urls=wrong_urls)


async def run_check_feeds() -> None:
    async with with_app():

        logger.info("start_experiment")

        sources = await execute("SELECT id, uid FROM f_sources")

        source_ids = {row["id"]: row["uid"] for row in sources}

        counter = 0
        wrong_urls = 0
        wrong_sources = 0

        feeds = await execute("SELECT url, source_id, uid FROM f_feeds")

        total = len(feeds)

        for row in feeds:
            counter += 1

            url = row["url"]
            uid = row["uid"]

            expected_uid = url_to_uid(url)
            expected_source_uid = url_to_source_uid(url)

            if uid != expected_uid:
                wrong_urls += 1

            if source_ids[row["source_id"]] != expected_source_uid:
                wrong_sources += 1

            if counter % 100 == 0:
                logger.info(
                    "processed_entries",
                    counter=counter,
                    percentage=round((counter / total) * 100, 2),
                    wrong_urls=wrong_urls,
                    wrong_sources=wrong_sources,
                )

        logger.info("experiment_finished", wrong_urls=wrong_urls, wrong_sources=wrong_sources)


async def run_experiment() -> None:
    pass


@cli_app.command()
def experiment() -> None:
    asyncio.run(run_experiment())



import uuid

import pytest
import pytest_asyncio

from ffun.domain.domain import new_collection_id
from ffun.domain.entities import CollectionSlug
from ffun.domain.urls import str_to_absolute_url, to_feed_url
from ffun.feeds_collections.collections import collections
from ffun.feeds_collections.entities import Collection, CollectionId, FeedInfo
from ffun.feeds_collections.settings import settings


@pytest.fixture(scope="session", autouse=True)
def collection_configs_must_be_none_in_tests() -> None:
    assert settings.collection_configs is None


@pytest_asyncio.fixture(scope="session")
async def collection_id_for_test_feeds(collection_configs_must_be_none_in_tests: None) -> CollectionId:

    feed_id_ = uuid.uuid4().hex

    feed = FeedInfo(
        url=to_feed_url(str_to_absolute_url(f"http://example.com/{feed_id_}")),
        title=f"Feed {feed_id_}",
        description=f"Feed {feed_id_}",
    )

    collection = Collection(
        id=new_collection_id(),
        slug=CollectionSlug(uuid.uuid4().hex),
        gui_order=1,
        name=uuid.uuid4().hex,
        description=uuid.uuid4().hex,
        feeds=[feed],
        show_on_main=True,
    )

    await collections.add_test_collection(collection)

    return collection.id



import uuid
from typing import Any

from ffun.core import utils
from ffun.domain.entities import AbsoluteUrl, UnknownUrl
from ffun.domain.urls import normalize_classic_unknown_url
from ffun.parsers.entities import EntryInfo


def fake_url() -> AbsoluteUrl:
    url = normalize_classic_unknown_url(UnknownUrl(f"https://{uuid.uuid4().hex}.com"))
    assert url is not None
    return url


def fake_title() -> str:
    return f"Entry Title: {uuid.uuid4().hex}"


def fake_body() -> str:
    return f"Entry Body: {uuid.uuid4().hex}"


def fake_entry_info(**kwargs: Any) -> EntryInfo:
    return EntryInfo(
        title=fake_title() if "title" not in kwargs else kwargs["title"],
        body=fake_body() if "body" not in kwargs else kwargs["body"],
        external_id=uuid.uuid4().hex if "external_id" not in kwargs else kwargs["external_id"],
        external_url=fake_url() if "external_url" not in kwargs else kwargs["external_url"],
        external_tags=(
            {uuid.uuid4().hex, uuid.uuid4().hex} if "external_tags" not in kwargs else kwargs["external_tags"]
        ),
        published_at=utils.now() if "published_at" not in kwargs else kwargs["published_at"],
    )
